import warnings

from pyramid.httpexceptions import HTTPFound

from kotti.interfaces import IContent
from kotti.testing import DummyRequest
from kotti.views.edit.default_views import DefaultViewSelection


class TestDefaultViewSelection:
    def test__is_valid_view(self, root, config):

        config.add_view(
            context=IContent,
            name="folder_view",
            permission="view",
            renderer="kotti:templates/view/folder.pt",
        )

        context = root
        request = DummyRequest()

        view = DefaultViewSelection(context, request)

        assert view._is_valid_view("folder_view") is True
        assert view._is_valid_view("foo") is False

    def test_default_views(self, root, config):
        config.add_view(
            context=IContent,
            name="folder_view",
            permission="view",
            renderer="kotti:templates/view/folder.pt",
        )

        context = root
        request = DummyRequest()

        view = DefaultViewSelection(context, request)

        sviews = view.default_view_selector()

        assert "selectable_default_views" in sviews

        # the root should have at least the default view and the folder_view
        assert len(sviews["selectable_default_views"]) > 1

        # the first view is always the default view
        assert sviews["selectable_default_views"][0]["is_current"] is True
        assert sviews["selectable_default_views"][0]["name"] == "default"
        assert sviews["selectable_default_views"][0]["title"] == "Default view"

        assert sviews["selectable_default_views"][1]["is_current"] is False

        # set the default view to folder_view view
        request = DummyRequest(GET={"view_name": "folder_view"})
        view = DefaultViewSelection(context, request)

        assert type(view.set_default_view()) == HTTPFound
        assert context.default_view == "folder_view"

        # set back to default
        request = DummyRequest(GET={"view_name": "default"})
        view = DefaultViewSelection(context, request)

        assert type(view.set_default_view()) == HTTPFound
        assert context.default_view is None

        # try to set non existing view
        request = DummyRequest(GET={"view_name": "nonexisting"})
        view = DefaultViewSelection(context, request)

        assert type(view.set_default_view()) == HTTPFound
        assert context.default_view is None

    def test_warning_for_non_registered_views(self, root):

        with warnings.catch_warnings(record=True) as w:

            DefaultViewSelection(root, DummyRequest()).default_view_selector()

            assert len(w) == 1
            assert issubclass(w[-1].category, UserWarning)
            assert (
                str(w[-1].message)
                == "No view called 'folder_view' is registered for <Document 1 at />."
            )  # noqa



from io import BytesIO

import pytest
from colander import null
from mock import MagicMock
from mock import patch
from pyramid.httpexceptions import HTTPMovedPermanently

from kotti.filedepot import StoredFileResponse
from kotti.testing import DummyRequest
from kotti.testing import asset
from kotti.views.file import attachment_view
from kotti.views.file import inline_view


class TestFileViews:
    def _create_file(self, config):
        from kotti.resources import File

        self.file = File(asset("logo.png").read(), "myfüle.png", "image/png")

    def _test_common_headers(self, headers):
        for name in ("Content-Disposition", "Content-Length", "Content-Type"):
            assert isinstance(headers[name], str)
        assert headers["Content-Length"] == "55715"
        assert headers["Content-Type"] == "image/png"

    @pytest.mark.parametrize(
        "params", [(inline_view, "inline"), (attachment_view, "attachment")]
    )
    def test_file_views(self, params, config, filedepot, dummy_request, depot_tween):
        view, disposition = params
        self._create_file(config)

        res = view(self.file, dummy_request)

        self._test_common_headers(res.headers)

        assert res.content_disposition.startswith(
            f"{disposition};filename=\"my"
        )

        data = asset("logo.png").read()
        assert res.body == data
        assert b"".join(res.app_iter) == data


class TestFileEditForm:
    def make_one(self):
        from kotti.views.edit.content import FileEditForm
        from kotti.resources import File

        return FileEditForm(File(), DummyRequest())

    def test_edit_with_file(self, db_session, filedepot):
        view = self.make_one()
        view.edit(
            title="A title",
            description="A description",
            tags=["A tag"],
            file=dict(
                fp=BytesIO(b"filecontents"),
                filename="myfile.png",
                mimetype="image/png",
                size=10,
                uid="randomabc",
            ),
        )
        assert view.context.title == "A title"
        assert view.context.description == "A description"
        assert view.context.data.file.read() == b"filecontents"
        assert view.context.filename == "myfile.png"
        assert view.context.mimetype == "image/png"
        assert view.context.size == len(b"filecontents")
        assert view.context.tags == ["A tag"]

    def test_edit_without_file(self, filedepot):
        view = self.make_one()
        view.context.data = b"filecontents"
        view.context.filename = "myfile.png"
        view.context.mimetype = "image/png"
        view.context.size = 777
        with patch("kotti.views.edit.content._to_fieldstorage") as tfs:
            view.edit(title="A title", description="A description", tags=[], file=null)
            assert tfs.call_count == 0
            assert view.context.title == "A title"
            assert view.context.description == "A description"
            assert view.context.data.file.read() == b"filecontents"
            assert view.context.filename == "myfile.png"
            assert view.context.mimetype == "image/png"
            assert view.context.size == 777


class TestFileAddForm:
    def make_one(self):
        from kotti.views.edit.content import FileAddForm

        return FileAddForm(MagicMock(), DummyRequest())

    def test_add(self, config, filedepot):
        view = self.make_one()
        file = view.add(
            title="A title",
            description="A description",
            tags=[],
            file=dict(
                fp=BytesIO(b"filecontents"),
                filename="myfile.png",
                mimetype="image/png",
                size=None,
                uid="randomabc",
            ),
        )

        assert file.title == "A title"
        assert file.description == "A description"
        assert file.tags == []
        assert file.data.file.read() == b"filecontents"
        assert file.filename == "myfile.png"
        assert file.mimetype == "image/png"
        assert file.size == len(b"filecontents")


class TestFileUploadTempStore:
    def make_one(self):
        from kotti.views.form import FileUploadTempStore

        return FileUploadTempStore(DummyRequest())

    def test_keys(self):
        tmpstore = self.make_one()
        tmpstore.session["important"] = 3
        tmpstore.session["_secret"] = 4
        assert tmpstore.keys() == ["important"]

    def test_delitem(self):
        tmpstore = self.make_one()
        tmpstore.session["important"] = 3
        del tmpstore["important"]
        assert "important" not in tmpstore.session

    def test_setitem_with_stream(self):
        ts = self.make_one()
        ts["a"] = {"fp": BytesIO(b"test"), "marker": "yes"}
        assert ts.session["a"] == {"file_contents": b"test", "marker": "yes"}
        v = ts["a"]
        assert "fp" in v.keys()
        assert v["marker"] == "yes"
        assert v["fp"].read() == b"test"

    def test_setitem_with_empty(self):
        ts = self.make_one()
        ts["a"] = {"fp": None, "marker": "yes"}
        assert ts.session["a"] == {"file_contents": b"", "marker": "yes"}
        assert ts["a"] == {"fp": None, "marker": "yes"}


class TestDepotStore:

    from kotti.resources import File

    @pytest.mark.parametrize("factory", [File])
    def test_create(self, factory, filedepot, image_asset, app):
        data = image_asset.read()
        f = factory(data)
        if factory.__class__.__name__ == "File":
            assert len(f.data["files"]) == 1
        elif factory.__class__.__name__ == "Image":
            assert len(f.data["files"]) == 4
        assert f.data.file.read() == data

    @pytest.mark.parametrize("factory", [File])
    def test_edit_content(
        self, factory, filedepot, image_asset, image_asset2, app, db_session
    ):
        data = image_asset.read()
        f = factory(data)
        assert f.data.file.read() == data
        db_session.flush()
        data2 = image_asset2.read()
        f.data = data2
        db_session.flush()

        assert f.data.file.read() == data2

    @pytest.mark.parametrize("factory", [File])
    def test_session_rollback(self, factory, db_session, filedepot, image_asset, app):
        storage = filedepot.get()

        f = factory(data=image_asset.read(), name="content", title="content")
        id = f.data["file_id"]

        db_session.add(f)
        db_session.flush()
        storage.get(id)

        db_session.rollback()
        with pytest.raises(IOError):
            storage.get(id)
        assert storage.delete.called

    @pytest.mark.parametrize("factory", [File])
    def test_delete(self, factory, db_session, root, filedepot, image_asset, app):

        storage = filedepot.get()

        f = factory(data=image_asset, name="content", title="content")
        id = f.data["file_id"]
        root[str(id)] = f
        db_session.flush()

        storage.get(id)

        del root[str(id)]
        import transaction

        transaction.commit()

        with pytest.raises(IOError):
            storage.get(id)
        assert storage.delete.called


class TestUploadedFileResponse:
    def _create_file(
        self, data=b"file contents", filename="myfüle.png", mimetype="image/png"
    ):
        from kotti.resources import File

        return File(data, filename, mimetype)

    def test_as_body(self, filedepot, image_asset, dummy_request):
        data = image_asset.read()
        f = self._create_file(data)
        # resp = UploadedFileResponse(f.data, DummyRequest())
        resp = dummy_request.uploaded_file_response(f.data)
        assert resp.body == data

    def test_as_app_iter(self, filedepot, image_asset, dummy_request):
        from pyramid.response import FileIter

        data = image_asset.read()
        f = self._create_file(data)
        # resp = UploadedFileResponse(f.data, DummyRequest())
        resp = dummy_request.uploaded_file_response(f.data)
        assert isinstance(resp.app_iter, FileIter)
        assert b"".join(resp.app_iter) == data

    def test_unknown_filename(self, filedepot, image_asset2, dummy_request):
        f = self._create_file(b"foo", "file.bar", None)
        # resp = UploadedFileResponse(f.data, DummyRequest())
        resp = dummy_request.uploaded_file_response(f.data)
        assert resp.headers["Content-Type"] == "application/octet-stream"

    def test_guess_content_type(self, filedepot, image_asset, dummy_request):
        f = self._create_file(image_asset.read(), "file.png", None)
        # resp = UploadedFileResponse(f.data, DummyRequest())
        resp = dummy_request.uploaded_file_response(f.data)
        assert resp.headers["Content-Type"] == "image/png"

    def test_caching(self, filedepot, monkeypatch, dummy_request):
        import datetime
        import webob.response

        f = self._create_file()
        d = datetime.datetime(2012, 12, 31, 13)

        class mockdatetime:
            @staticmethod
            def utcnow():
                return d

        monkeypatch.setattr(webob.response, "datetime", mockdatetime)

        # resp = UploadedFileResponse(f.data, DummyRequest(), cache_max_age=10)
        resp = dummy_request.uploaded_file_response(f.data, cache_max_age=10)

        # this is set by filedepot fixture
        assert resp.headers["Last-Modified"] == "Sun, 30 Dec 2012 00:00:00 GMT"
        assert resp.headers["Expires"] == "Mon, 31 Dec 2012 13:00:10 GMT"

    def test_redirect(self, filedepot, dummy_request):
        class PublicFile:
            public_url = "http://example.com"

        class PublicData:
            file = PublicFile()

        with pytest.raises(HTTPMovedPermanently) as e:
            # UploadedFileResponse(PublicData(), DummyRequest())
            dummy_request.uploaded_file_response(PublicData())

        response = e.value
        assert response.headers["Location"] == "http://example.com"


class TestStoredFileResponse:
    def _create_file(
        self, data=b"file contents", filename="myfüle.png", mimetype="image/png"
    ):
        from kotti.resources import File

        return File(data, filename, mimetype)

    def test_as_body(self, filedepot, image_asset, dummy_request):
        data = image_asset.read()
        f = self._create_file(data)
        resp = StoredFileResponse(f.data.file, dummy_request)
        assert resp.body == data

    def test_as_app_iter(self, filedepot, image_asset, dummy_request):
        from pyramid.response import FileIter

        data = image_asset.read()
        f = self._create_file(data)
        resp = StoredFileResponse(f.data.file, dummy_request)
        assert isinstance(resp.app_iter, FileIter)
        assert b"".join(resp.app_iter) == data

    def test_unknown_filename(self, filedepot, image_asset2, dummy_request):
        f = self._create_file(b"foo", "file.bar", None)
        resp = StoredFileResponse(f.data.file, dummy_request)
        assert resp.headers["Content-Type"] == "application/octet-stream"

    def test_guess_content_type(self, filedepot, image_asset, dummy_request):
        f = self._create_file(image_asset.read(), "file.png", None)
        resp = StoredFileResponse(f.data.file, dummy_request)
        assert resp.headers["Content-Type"] == "image/png"

    def test_caching(self, filedepot, monkeypatch, dummy_request):
        import datetime
        import webob.response

        f = self._create_file()
        d = datetime.datetime(2012, 12, 31, 13)

        class mockdatetime:
            @staticmethod
            def utcnow():
                return d

        monkeypatch.setattr(webob.response, "datetime", mockdatetime)

        resp = dummy_request.uploaded_file_response(f.data)
        assert resp.headers["Expires"] == "Mon, 07 Jan 2013 13:00:00 GMT"

        resp = dummy_request.uploaded_file_response(f.data, cache_max_age=10)
        assert resp.headers["Expires"] == "Mon, 31 Dec 2012 13:00:10 GMT"

        # this is set by filedepot fixture
        assert resp.headers["Last-Modified"] == "Sun, 30 Dec 2012 00:00:00 GMT"

    def test_redirect(self, filedepot, dummy_request):
        class PublicFile:
            public_url = "http://example.com"

        with pytest.raises(HTTPMovedPermanently) as e:
            StoredFileResponse(PublicFile(), dummy_request)

        response = e.value
        assert response.headers["Location"] == "http://example.com"



from pyramid.view import view_config

from kotti.interfaces import IFile


@view_config(
    name="view",
    context=IFile,
    permission="view",
    renderer="kotti:templates/view/file.pt",
)
def view(context, request):
    return {}


@view_config(name="inline-view", context=IFile, permission="view")
def inline_view(context, request):
    return request.uploaded_file_response(context.data)


@view_config(name="attachment-view", context=IFile, permission="view")
def attachment_view(context, request):
    return request.uploaded_file_response(context.data, "attachment")


def includeme(config):
    """ Pyramid includeme hook.

    :param config: app config
    :type config: :class:`pyramid.config.Configurator`
    """

    config.scan(__name__)



"""
:summary: Default view selctor views
"""
import warnings

from pyramid.compat import map_
from pyramid.httpexceptions import HTTPFound
from pyramid.interfaces import IView
from pyramid.interfaces import IViewClassifier
from pyramid.threadlocal import get_current_registry
from pyramid.view import view_config
from pyramid.view import view_defaults
from zope.interface import providedBy

from kotti.util import _


@view_defaults(permission="edit")
class DefaultViewSelection:
    def __init__(self, context, request):

        self.context = context
        self.request = request

    def _get_view(self, view_name):  # pragma: no cover
        """This code is copied from pyramid.view.
           We trust it and don't test.

           Returns True if a view with name view_name is registered for context.
        """
        provides = [IViewClassifier] + map_(providedBy, (self.request, self.context))

        try:
            reg = self.request.registry
        except AttributeError:
            reg = get_current_registry()

        return reg.adapters.lookup(provides, IView, name=view_name)

    def _is_valid_view(self, view_name):
        """Return True if a view with name view_name is registered for context.
        """

        return self._get_view(view_name) is not None

    def _is_valid_default_view(self, view_name):
        """Return True if a view with name view_name is a valid choice as
        default view for context.
        """

        return self._is_valid_view(view_name) and view_name in [
            v[0] for v in self.context.type_info.selectable_default_views
        ]

    @view_config(
        name="default-view-selector",
        renderer="kotti:templates/default-view-selector.pt",
    )
    def default_view_selector(self):
        """
        :summary: Submenu for selection of the node's default view.
        :result: Dictionary with a selectable_default_views list.
        :rtype: dict
        """

        sviews = []

        for v in self.context.type_info.selectable_default_views:
            name, title = v
            if self._is_valid_view(name):
                sviews.append(
                    {
                        "name": name,
                        "title": title,
                        "is_current": name == self.context.default_view,
                    }
                )
            else:
                warnings.warn(
                    "No view called '{}' is registered for {!r}.".format(
                        name, self.context
                    )
                )

        return {
            "selectable_default_views": [
                {
                    "name": "default",
                    "title": _("Default view"),
                    "is_current": self.context.default_view is None,
                }
            ]
            + sviews
        }

    @view_config(name="set-default-view")
    def set_default_view(self):
        """
        :summary: Set the node's default view and redirect to it.
        :result: Redirect to the context URL.
        :rtype: pyramid.httpexceptions.HTTPFound
        """

        if "view_name" in self.request.GET:

            view_name = self.request.GET["view_name"]

            if view_name == "default":
                self.context.default_view = None
                self.request.session.flash(
                    _("Default view has been reset to default."), "success"
                )
            else:
                if self._is_valid_default_view(view_name):
                    self.context.default_view = view_name
                    self.request.session.flash(
                        _("Default view has been set."), "success"
                    )
                else:
                    self.request.session.flash(
                        _("Default view could not be set."), "error"
                    )

        return HTTPFound(location=self.request.resource_url(self.context))


def includeme(config):
    """ Pyramid includeme hook.

    :param config: app config
    :type config: :class:`pyramid.config.Configurator`
    """

    config.scan(".default_views")



from mock import patch
from pyramid.authentication import CallbackAuthenticationPolicy
from pytest import raises

from kotti.testing import DummyRequest


class HasPermissionTests:
    def test_has_permission(self):
        import mock
        from kotti.security import has_permission

        permission = "edit"
        context = object()
        request = mock.MagicMock()
        has_permission(permission, context, request)
        assert (
            request.has_permission.assert_called_once_with(permission, context=context)
            is None
        )


class TestGroups:
    def test_root_default(self, db_session, root):
        from kotti.security import list_groups
        from kotti.security import list_groups_raw

        assert list_groups("admin", root) == ["role:admin"]
        assert list_groups_raw("admin", root) == set()

    def test_empty(self, db_session, root):
        from kotti.security import list_groups

        assert list_groups("bob", root) == []

    def test_simple(self, db_session, root):
        from kotti.security import list_groups
        from kotti.security import list_groups_raw
        from kotti.security import set_groups

        set_groups("bob", root, ["role:editor"])
        assert list_groups("bob", root) == ["role:editor"]
        assert list_groups_raw("bob", root) == {"role:editor"}

    def test_not_a_node(self):
        from kotti.security import list_groups_raw

        assert list_groups_raw("bob", object()) == set()

    def test_overwrite_and_delete(self, db_session, root):
        from kotti.resources import get_root
        from kotti.security import list_groups
        from kotti.security import list_groups_raw
        from kotti.security import set_groups

        set_groups("bob", root, ["role:editor"])
        assert list_groups("bob", root) == ["role:editor"]
        assert list_groups_raw("bob", root) == {"role:editor"}

        set_groups("bob", root, ["role:admin"])
        assert list_groups("bob", root) == ["role:admin"]
        assert list_groups_raw("bob", root) == {"role:admin"}

        set_groups("bob", root)
        assert list_groups("bob", root) == []
        assert get_root() is root

    def test_inherit(self, db_session, root):
        from kotti.resources import Node
        from kotti.security import list_groups
        from kotti.security import list_groups_raw
        from kotti.security import set_groups

        child = root["child"] = Node()
        db_session.flush()

        assert list_groups("bob", child) == []
        set_groups("bob", root, ["role:editor"])
        assert list_groups("bob", child) == ["role:editor"]

        # Groups from the child are added:
        set_groups("bob", child, ["group:somegroup"])
        assert set(list_groups("bob", child)) == {"group:somegroup", "role:editor"}

        # We can ask to list only those groups that are defined locally:
        assert list_groups_raw("bob", child) == {"group:somegroup"}

    @staticmethod
    def add_some_groups(db_session, root):
        from kotti.resources import Node
        from kotti.security import set_groups

        child = root["child"] = Node()
        grandchild = child["grandchild"] = Node()
        db_session.flush()

        # root:
        #   bob               -> group:bobsgroup
        #   frank             -> group:franksgroup
        #   group:franksgroup -> role:editor
        # child:
        #   group:bobsgroup   -> group:franksgroup
        # grandchild:
        #   group:franksgroup -> role:admin
        #   group:franksgroup -> group:bobsgroup

        # bob and frank are a site-wide members of their respective groups:
        set_groups("bob", root, ["group:bobsgroup"])
        set_groups("frank", root, ["group:franksgroup"])

        # franksgroup has a site-wide editor role:
        set_groups("group:franksgroup", root, ["role:editor"])

        # bobsgroup is part of franksgroup on the child level:
        set_groups("group:bobsgroup", child, ["group:franksgroup"])

        # franksgroup has the admin role on the grandchild.
        # and finally, to test recursion, we make franksgroup part of
        # bobsgroup on the grandchild level:
        set_groups("group:franksgroup", grandchild, ["role:owner", "group:bobsgroup"])

    def test_nested_groups(self, db_session, root):
        from kotti.security import list_groups
        from kotti.security import list_groups_ext

        self.add_some_groups(db_session, root)
        child = root["child"]
        grandchild = child["grandchild"]

        # Check bob's groups on every level:
        assert list_groups("bob", root) == ["group:bobsgroup"]
        assert set(list_groups("bob", child)) == {
            "group:bobsgroup",
            "group:franksgroup",
            "role:editor",
        }
        assert set(list_groups("bob", grandchild)) == {
            "group:bobsgroup",
            "group:franksgroup",
            "role:editor",
            "role:owner",
        }

        # Check group:franksgroup groups on every level:
        assert set(list_groups("frank", root)) == {"group:franksgroup", "role:editor"}
        assert set(list_groups("frank", child)) == {"group:franksgroup", "role:editor"}
        assert set(list_groups("frank", grandchild)) == {
            "group:franksgroup",
            "role:editor",
            "role:owner",
            "group:bobsgroup",
        }

        # Sometimes it's useful to know which of the groups were
        # inherited, that's what 'list_groups_ext' is for:
        groups, inherited = list_groups_ext("bob", root)
        assert groups == ["group:bobsgroup"]
        assert inherited == []

        groups, inherited = list_groups_ext("bob", child)
        assert set(groups) == {"group:bobsgroup", "group:franksgroup", "role:editor"}
        assert set(inherited) == {"group:bobsgroup", "group:franksgroup", "role:editor"}

        groups, inherited = list_groups_ext("group:bobsgroup", child)
        assert set(groups) == {"group:franksgroup", "role:editor"}
        assert inherited == ["role:editor"]

        groups, inherited = list_groups_ext("group:franksgroup", grandchild)
        assert set(groups) == {"group:bobsgroup", "role:owner", "role:editor"}
        assert inherited == ["role:editor"]

    def test_works_with_auth(self, db_session, root):
        from kotti.resources import Node
        from kotti.security import get_principals
        from kotti.security import list_groups_callback
        from kotti.security import set_groups

        child = root["child"] = Node()
        db_session.flush()

        request = DummyRequest()
        auth = CallbackAuthenticationPolicy()
        auth.unauthenticated_userid = lambda *args: "bob"
        auth.callback = list_groups_callback

        request.context = root
        assert auth.effective_principals(request) == [  # user doesn't exist yet
            "system.Everyone"
        ]

        get_principals()["bob"] = dict(name="bob")
        assert auth.effective_principals(request) == [
            "system.Everyone",
            "system.Authenticated",
            "bob",
        ]

        # Define that bob belongs to bobsgroup on the root level:
        set_groups("bob", root, ["group:bobsgroup"])
        request.context = child
        assert set(auth.effective_principals(request)) == {
            "system.Everyone",
            "system.Authenticated",
            "bob",
            "group:bobsgroup",
        }

        # define that bob belongs to franksgroup in the user db:
        get_principals()["bob"].groups = ["group:franksgroup"]
        set_groups("group:franksgroup", child, ["group:anothergroup"])
        assert set(auth.effective_principals(request)) == {
            "system.Everyone",
            "system.Authenticated",
            "bob",
            "group:bobsgroup",
            "group:franksgroup",
            "group:anothergroup",
        }

        # And lastly test that circular group defintions are not a
        # problem here either:
        get_principals()["group:franksgroup"] = dict(
            name="group:franksgroup",
            title="Frank's group",
            groups=["group:funnygroup", "group:bobsgroup"],
        )
        assert set(auth.effective_principals(request)) == {
            "system.Everyone",
            "system.Authenticated",
            "bob",
            "group:bobsgroup",
            "group:franksgroup",
            "group:anothergroup",
            "group:funnygroup",
        }

    def test_list_groups_callback_with_groups(self, db_session):
        from kotti.security import list_groups_callback
        from kotti.security import get_principals

        # Although group definitions are also in the user database,
        # we're not allowed to authenticate with a group id:
        get_principals()["bob"] = dict(name="bob")
        get_principals()["group:bobsgroup"] = dict(name="group:bobsgroup")

        request = DummyRequest()
        assert list_groups_callback("bob", request) == []
        assert list_groups_callback("group:bobsgroup", request) is None

    def test_principals_with_local_roles(self, db_session, root):
        from kotti.resources import Node
        from kotti.security import map_principals_with_local_roles
        from kotti.security import principals_with_local_roles
        from kotti.security import set_groups

        child = root["child"] = Node()
        db_session.flush()

        assert principals_with_local_roles(root) == []
        assert principals_with_local_roles(child) == []
        assert map_principals_with_local_roles(root) == []
        assert map_principals_with_local_roles(child) == []

        set_groups("group:bobsgroup", child, ["role:editor"])
        set_groups("bob", root, ["group:bobsgroup"])
        set_groups("group:franksgroup", root, ["role:editor"])

        assert set(principals_with_local_roles(child)) == {
            "bob",
            "group:bobsgroup",
            "group:franksgroup",
        }
        assert set(principals_with_local_roles(child, inherit=False)) == {
            "group:bobsgroup"
        }
        assert set(principals_with_local_roles(root)) == {"bob", "group:franksgroup"}

    def test_copy_local_groups(self, db_session, root):
        from kotti.security import principals_with_local_roles
        from kotti.security import set_groups

        self.test_principals_with_local_roles(db_session, root)
        child = root["child"]
        assert set(principals_with_local_roles(child)) == {
            "bob",
            "group:bobsgroup",
            "group:franksgroup",
        }

        # We make a copy of 'child', and we expect the local roles set
        # on 'child' to not be copied over:
        child2 = root["child2"] = child.copy()
        db_session.flush()
        assert set(principals_with_local_roles(child2)) == {"bob", "group:franksgroup"}
        assert len(principals_with_local_roles(child)) == 3

        # When we now change the local roles of 'child', the copy is
        # unaffected:
        set_groups("group:bobsgroup", child, [])
        assert len(principals_with_local_roles(child)) == 2
        assert len(principals_with_local_roles(child2)) == 2

    def test_map_principals_with_local_roles(self, db_session, root):
        from kotti.security import get_principals
        from kotti.security import map_principals_with_local_roles

        self.test_principals_with_local_roles(db_session, root)
        child = root["child"]
        P = get_principals()

        # No users are defined in P, thus we get the empty list:
        assert map_principals_with_local_roles(root) == []

        P["bob"] = {"name": "bob"}
        P["group:bobsgroup"] = {"name": "group:bobsgroup"}

        value = map_principals_with_local_roles(root)
        assert len(value) == 1
        bob, (bob_all, bob_inherited) = value[0]
        assert bob_all == ["group:bobsgroup"]
        assert bob_inherited == []

        value = map_principals_with_local_roles(child)
        assert len(value) == 2
        bob, (bob_all, bob_inherited) = value[0]
        bobsgroup, (bobsgroup_all, bobsgroup_inherited) = value[1]
        assert set(bob_all) == {"group:bobsgroup", "role:editor"}
        assert set(bob_inherited) == {"group:bobsgroup", "role:editor"}
        assert bobsgroup_all == ["role:editor"]
        assert bobsgroup_inherited == []

    def test_local_roles_db_cascade(self, db_session, root):
        from kotti.resources import LocalGroup
        from kotti.resources import Node
        from kotti.security import set_groups

        child = root["child"] = Node()
        db_session.flush()

        # We set a local group on child and delete child.  We then
        # expect the LocalGroup entry to have been deleted from the
        # database:
        assert db_session.query(LocalGroup).count() == 0
        set_groups("group:bobsgroup", child, ["role:editor"])
        assert db_session.query(LocalGroup).count() == 1
        del root["child"]
        db_session.flush()
        assert db_session.query(LocalGroup).count() == 0


class TestPrincipals:
    def get_principals(self):
        from kotti.security import get_principals

        return get_principals()

    def make_bob(self):
        users = self.get_principals()
        users["bob"] = dict(
            name="bob",
            password="secret",
            email="bob@dabolina.com",
            title="Bob Dabolina",
            groups=["group:bobsgroup"],
        )
        return users["bob"]

    def _assert_is_bob(self, bob):
        assert bob.name == "bob"
        assert bob.title == "Bob Dabolina"
        assert bob.groups == ["group:bobsgroup"]

    def test_hash_password_non_ascii(self, db_session):
        self.get_principals().hash_password("\xd6TEst")

    def test_default_admin(self, db_session):
        admin = self.get_principals()["admin"]
        assert self.get_principals().validate_password("secret", admin.password)
        assert admin.groups == ["role:admin"]

    def test_users_empty(self, db_session):
        users = self.get_principals()
        with raises(KeyError):
            users["bob"]
        with raises(KeyError):
            del users["bob"]
        assert users.keys() == ["admin"]

    def test_users_add_and_remove(self, db_session):
        self.make_bob()
        users = self.get_principals()
        self._assert_is_bob(users["bob"])
        assert set(users.keys()) == {"admin", "bob"}

        del users["bob"]
        with raises(KeyError):
            users["bob"]
        with raises(KeyError):
            del users["bob"]

    def test_users_search(self, db_session):
        users = self.get_principals()
        assert list(users.search(name="*Bob*")) == []
        self.make_bob()
        [bob] = list(users.search(name="bob"))
        self._assert_is_bob(bob)
        [bob] = list(users.search(name="*Bob*"))
        self._assert_is_bob(bob)
        [bob] = list(users.search(name="*Bob*", title="*Frank*"))
        self._assert_is_bob(bob)
        with raises(AttributeError):
            users.search(name="bob", foo="bar")
        assert list(users.search()) == []
        # search on a non string attribute
        assert bob in list(users.search(active=True))
        # test operator
        assert len(list(users.search(name="bob", active=True))) == 2
        assert len(list(users.search(match="all", name="bob", active=True))) == 1

    def test_groups_from_users(self, db_session, root):
        from kotti.resources import Node
        from kotti.security import list_groups
        from kotti.security import set_groups

        self.make_bob()
        child = root["child"] = Node()
        db_session.flush()

        assert list_groups("bob", root) == ["group:bobsgroup"]

        set_groups("group:bobsgroup", root, ["role:editor"])
        set_groups("role:editor", child, ["group:foogroup"])

        assert set(list_groups("bob", root)) == {"group:bobsgroup", "role:editor"}
        assert set(list_groups("bob", child)) == {
            "group:bobsgroup",
            "role:editor",
            "group:foogroup",
        }

    def test_is_user(self, db_session):
        from kotti.security import is_user

        bob = self.make_bob()
        assert is_user(bob) is True
        bob.name = "group:bobsgroup"
        assert is_user(bob) is False

    def test_hash_and_validate_password(self, db_session):
        password = "secret"
        principals = self.get_principals()
        hashed = principals.hash_password(password)
        assert principals.validate_password(password, hashed)
        assert principals.validate_password("foo", hashed) is False

    def test_bobs_hashed_password(self, db_session):
        bob = self.make_bob()
        principals = self.get_principals()
        assert principals.validate_password("secret", bob.password)

        # When we set the 'password' attribute directly, we have to do
        # the hashing ourselves.  This won't work:
        bob.password = "anothersecret"
        assert principals.validate_password("anothersecret", bob.password) is False

        # This will:
        bob.password = principals.hash_password("anothersecret")
        assert principals.validate_password("anothersecret", bob.password)

    def test_active(self, db_session):
        bob = self.make_bob()
        assert bob.active is True
        bob.active = False
        assert bob.active is False

    def test_reset_password(self, db_session):
        from kotti.views.login import login

        request = DummyRequest()
        self.make_bob()
        request.params["reset-password"] = "on"
        request.params["login"] = "bob"
        request.params["password"] = "secret"
        with patch("kotti.views.login.email_set_password") as email_set_password:
            login(None, request)
        assert request.session.pop_flash("success") == [
            "You should be receiving an email with a link to reset your "
            "password. Doing so will activate your account."
        ]
        assert email_set_password.call_count == 1

    def test_reset_password_inactive_user(self, db_session):
        from kotti.views.login import login

        request = DummyRequest()
        self.make_bob().active = False
        request.params["reset-password"] = "on"
        request.params["login"] = "bob"
        request.params["password"] = "secret"
        with patch("kotti.views.login.email_set_password") as email_set_password:
            login(None, request)
        assert request.session.pop_flash("error") == [
            "That username or email is not known by this system."
        ]
        assert email_set_password.call_count == 0

    def test_login(self, db_session):
        from kotti.views.login import login

        request = DummyRequest()

        # No login attempt:
        result = login(None, request)
        assert isinstance(result, dict)
        assert request.session.pop_flash("success") == []
        assert request.session.pop_flash("error") == []

        # Attempt to log in before Bob exists:
        request.params["submit"] = "on"
        request.params["login"] = "bob"
        request.params["password"] = "secret"
        result = login(None, request)
        assert isinstance(result, dict)
        assert request.session.pop_flash("success") == []
        assert request.session.pop_flash("error") == ["Login failed."]

        # Make Bob and do it again:
        bob = self.make_bob()
        assert bob.last_login_date is None
        result = login(None, request)
        assert result.status == "302 Found"
        assert [request.session.pop_flash("success")[0].interpolate()] == [
            "Welcome, Bob Dabolina!"
        ]
        last_login_date = bob.last_login_date
        assert last_login_date is not None
        assert request.session.pop_flash("error") == []

        # Log in with email:
        request.params["login"] = "bob@dabolina.com"
        result = login(None, request)
        assert result.status == "302 Found"
        assert [request.session.pop_flash("success")[0].interpolate()] == [
            "Welcome, Bob Dabolina!"
        ]
        assert last_login_date < bob.last_login_date

        # Deactive Bob, logging in is no longer possible:
        bob.active = False
        result = login(None, request)
        assert isinstance(result, dict)
        assert request.session.pop_flash("error") == ["Login failed."]

        # If Bob has a 'confirm_token' set, logging in is still possible:
        bob.active = True
        bob.confirm_token = "token"
        result = login(None, request)
        assert result.status == "302 Found"
        assert [request.session.pop_flash("success")[0].interpolate()] == [
            "Welcome, Bob Dabolina!"
        ]

    def test_login_with_email_remembers(self, db_session):
        from kotti.views.login import login

        request = DummyRequest()

        self.make_bob()
        request.params["submit"] = "on"
        request.params["login"] = "bob@dabolina.com"
        request.params["password"] = "secret"
        with patch("kotti.views.login.remember") as remember:
            login(None, request)
            remember.assert_called_with(request, "bob")


class TestAuthzContextManager:
    def test_basic(self):
        from kotti.security import authz_context

        context, request = object(), DummyRequest()
        with authz_context(context, request):
            assert request.environ["authz_context"] == context
        assert "authz_context" not in request.environ

    def test_set_before(self):
        from kotti.security import authz_context

        context, context2, request = object(), object(), DummyRequest()
        request.environ["authz_context"] = context2
        with authz_context(context, request):
            assert request.environ["authz_context"] == context
        assert request.environ["authz_context"] == context2

    def test_with_exception(self):
        from kotti.security import authz_context

        context, context2, request = object(), object(), DummyRequest()
        request.environ["authz_context"] = context2
        try:
            with authz_context(context, request):
                assert request.environ["authz_context"] == context
                raise ValueError
        except ValueError:
            assert request.environ["authz_context"] == context2


class TestViewPermitted:
    @patch("kotti.security.view_execution_permitted")
    def test_with_post_request(self, view_execution_permitted):
        from kotti.security import view_permitted

        context, request = object(), DummyRequest()
        request.method = "POST"
        calls = []

        def view_execution_permitted_mock(context, request, name):
            calls.append((context, request, name))
            assert request.method == "GET"

        view_execution_permitted.side_effect = view_execution_permitted_mock
        view_permitted(context, request)
        assert len(calls) == 1
        assert request.method == "POST"


class TestHasPermission:
    def test_request_has_permission(self):

        from kotti.request import Request

        permission, context = object(), object()
        args = []

        def has_permission_fake(self, permission, context=None):
            args.append((permission, context))
            assert self.environ["authz_context"] == context

        with patch(
            "kotti.request.pyramid.request.Request.has_permission",
            new=has_permission_fake,
        ):
            request = Request.blank("/")
            request.has_permission(permission, context)

        assert args == [(permission, context)]


class TestRolesSetters:
    def test_set_roles(self):
        from kotti.security import ROLES
        from kotti.security import set_roles
        from kotti.security import reset_roles

        before = ROLES.copy()
        set_roles({"role:admin": ROLES["role:admin"]})
        assert ROLES == {"role:admin": ROLES["role:admin"]}
        reset_roles()
        assert ROLES == before

    def test_set_sharing_roles(self):
        from kotti.security import SHARING_ROLES
        from kotti.security import set_sharing_roles
        from kotti.security import reset_sharing_roles

        before = SHARING_ROLES[:]
        set_sharing_roles(["role:admin"])
        assert SHARING_ROLES == ["role:admin"]
        reset_sharing_roles()
        assert SHARING_ROLES == before

    def test_set_user_management_roles(self):
        from kotti.security import USER_MANAGEMENT_ROLES
        from kotti.security import set_user_management_roles
        from kotti.security import reset_user_management_roles

        before = USER_MANAGEMENT_ROLES[:]
        set_user_management_roles(["role:admin"])
        assert USER_MANAGEMENT_ROLES == ["role:admin"]
        reset_user_management_roles()
        assert USER_MANAGEMENT_ROLES == before



class TestMassageGroups:
    def test_massage_groups_in(self):
        from kotti.views.users import _massage_groups_in

        appstruct = {"groups": ["user-group"], "roles": ["role:editor"]}
        _massage_groups_in(appstruct)
        assert appstruct["groups"] == ["role:editor", "group:user-group"]
        assert "roles" not in appstruct

        appstruct = {}
        _massage_groups_in(appstruct)
        assert appstruct["groups"] == []
        assert "roles" not in appstruct

    def test_massage_groups_out(self):
        from kotti.views.users import _massage_groups_out

        appstruct = {"groups": ["group:user-group", "role:editor"]}
        _massage_groups_out(appstruct)
        assert appstruct["groups"] == ["user-group"]
        assert appstruct["roles"] == ["role:editor"]

        appstruct = {"groups": []}
        _massage_groups_out(appstruct)
        assert appstruct["groups"] == []
        assert appstruct["roles"] == []



import mock  # must not ``from mock import call``, causes memory error.

from kotti.testing import DummyRequest


class TestRegister:
    def test_register_form(self, root):
        from kotti.views.login import register

        request = DummyRequest()
        res = register(root, request)
        assert res["form"][:5] == "<form"

    def test_register_submit_empty(self, root):
        from kotti.views.login import register

        request = DummyRequest()
        request.POST["register"] = "register"
        res = register(root, request)
        assert "There was a problem with your submission" in res["form"]

    def test_register_submit(self, root):
        from kotti.views.login import register
        from pyramid.httpexceptions import HTTPFound

        request = DummyRequest()
        request.POST["title"] = "Test User"
        request.POST["name"] = "test"
        request.POST["email"] = "test@example.com"
        request.POST["register"] = ("register",)

        with mock.patch("kotti.views.login.UserAddFormView") as form:
            with mock.patch("kotti.views.login.get_principals"):
                res = register(root, request)
                form.assert_has_calls(
                    [
                        mock.call().add_user_success(
                            {
                                "name": "test",
                                "roles": "",
                                "title": "Test User",
                                "send_email": True,
                                "groups": "",
                                "email": "test@example.com",
                            }
                        )
                    ]
                )
        assert isinstance(res, HTTPFound)

    def test_register_event(self, root):
        from kotti.views.login import register

        request = DummyRequest()
        request.POST["title"] = "Test User"
        request.POST["name"] = "test"
        request.POST["email"] = "test@example.com"
        request.POST["register"] = ("register",)

        with mock.patch("kotti.views.login.UserAddFormView"):
            with mock.patch("kotti.views.login.get_principals"):
                with mock.patch("kotti.views.login.notify") as notify:
                    register(root, request)
        assert notify.call_count == 1

    def test_register_submit_groups_and_roles(self, root):
        from kotti.views.login import register
        from pyramid.httpexceptions import HTTPFound

        request = DummyRequest()
        request.POST["title"] = "Test User"
        request.POST["name"] = "test"
        request.POST["email"] = "test@example.com"
        request.POST["register"] = ("register",)

        with mock.patch("kotti.views.login.UserAddFormView") as form:
            with mock.patch("kotti.views.login.get_principals"):
                with mock.patch(
                    "kotti.views.login.get_settings"
                ) as get_settings:  # noqa
                    get_settings.return_value = {
                        "kotti.register.group": "mygroup",
                        "kotti.register.role": "myrole",
                    }

                    res = register(root, request)

        form.assert_has_calls(
            [
                mock.call().add_user_success(
                    {
                        "name": "test",
                        "roles": {"role:myrole"},
                        "title": "Test User",
                        "send_email": True,
                        "groups": ["mygroup"],
                        "email": "test@example.com",
                    }
                )
            ]
        )
        assert isinstance(res, HTTPFound)


class TestNotRegister:
    def test_it(self, app):
        res = app.post("/register", status=404)
        assert res.status == "404 Not Found"



import colander
from mock import MagicMock
from mock import patch

from kotti.testing import Dummy
from kotti.testing import DummyRequest


class TestBaseFormView:
    def make(self):
        from kotti.views.form import BaseFormView

        return BaseFormView(Dummy(), DummyRequest())

    def test_init(self):
        from kotti.views.form import BaseFormView

        view = BaseFormView(Dummy(), DummyRequest(), more="args")
        assert view.more == "args"

    def test_schema_factory_bind(self):
        view = self.make()
        schema = MagicMock()
        view.schema_factory = lambda: schema
        view.__call__()
        assert view.schema == schema.bind.return_value
        schema.bind.assert_called_with(request=view.request)

    def test_use_csrf_token(self):
        view = self.make()
        schema = view.schema = MagicMock()
        view.__call__()
        assert schema.children.append.called
        assert schema.children.append.call_args[0][0].name == "csrf_token"

    def test_use_csrf_token_not(self):
        view = self.make()
        view.use_csrf_token = False
        schema = view.schema = MagicMock()
        view.__call__()
        assert not schema.children.append.called


class TestGetAppstruct:
    def call(self, *args, **kwargs):
        from kotti.views.form import get_appstruct

        return get_appstruct(*args, **kwargs)

    def test_schema_has_more_children_than_attrs(self):
        context = Dummy(first="firstvalue")
        schema = Dummy(children=[Dummy(name="first"), Dummy(name="second")])
        assert self.call(context, schema) == {"first": "firstvalue"}

    def test_schema_has_fewer_children_than_attrs(self):
        context = Dummy(first="firstvalue", second="secondvalue")
        schema = Dummy(children=[Dummy(name="first")])
        assert self.call(context, schema) == {"first": "firstvalue"}

    def test_none_converted_to_null(self):
        context = Dummy(first=None)
        schema = Dummy(children=[Dummy(name="first")])
        assert self.call(context, schema) == {"first": colander.null}


class TestEditFormView:
    def make(self):
        from kotti.views.form import EditFormView

        return EditFormView(Dummy(), DummyRequest(), success_url="http://localhost")

    def test_before(self):
        view = self.make()
        view.context.three = 3
        view.schema = Dummy()
        view.schema.children = [Dummy(name="three")]
        form = Dummy()
        view.before(form)
        assert form.appstruct == {"three": 3}

    def test_save_success_calls_edit(self):
        view = self.make()
        view.edit = MagicMock()
        view.save_success({"three": 3})
        view.edit.assert_called_with(three=3)

    def test_save_success_redirects(self):
        view = self.make()
        result = view.save_success({"three": 3})
        assert result.status == "302 Found"
        assert result.location == view.success_url

    def test_save_success_redirects_to_resource_url(self):
        view = self.make()
        view.success_url = None
        view.request.resource_url = lambda context: context.__class__.__name__
        result = view.save_success({"three": 3})
        assert result.status == "302 Found"
        assert result.location == "Dummy"


class TestAddFormView:
    def make(self):
        from kotti.views.form import AddFormView

        return AddFormView(Dummy(), DummyRequest())

    def test_save_success_calls_add(self):
        view = self.make()
        view.add = MagicMock()
        view.find_name = lambda appstruct: "somename"
        view.request.resource_url = lambda context: ""
        view.save_success({"three": 3})
        view.add.assert_called_with(three=3)
        assert view.add.return_value == view.context["somename"]

    def test_save_success_redirects(self):
        view = self.make()
        view.add = MagicMock()
        view.find_name = lambda appstruct: "somename"
        view.request.resource_url = lambda context: "MagicMock"
        result = view.save_success({"three": 3})
        assert result.status == "302 Found"
        assert result.location == "MagicMock"

    def test_save_success_redirects_custom_url(self):
        view = self.make()
        view.add = MagicMock()
        view.success_url = "there"
        view.find_name = lambda appstruct: "somename"
        result = view.save_success({"three": 3})
        assert result.status == "302 Found"
        assert result.location == "there"

    @patch("kotti.views.form.title_to_name")
    def test_find_name_uses_title_to_name(self, title_to_name):
        view = self.make()
        title_to_name.return_value = "cafe"
        assert view.find_name({"title": "Bar"}) == "cafe"
        title_to_name.assert_called_with("Bar", blacklist={}.keys())



from typing import Optional
from typing import Union

import pyramid.request
from pyramid.decorator import reify
from pyramid.interfaces import IRequest
from pyramid.security import Allowed
from pyramid.security import Denied
from zope.interface import implementer

from kotti.security import Principal
from kotti.security import get_user


@implementer(IRequest)
class Request(pyramid.request.Request):
    """ Kotti subclasses :class:`pyramid.request.Request` to make additional
    attributes / methods available on request objects and override Pyramid's
    :meth:`pyramid.request.Request.has_permission`.  The latter is needed to
    support Kotti's concept of local roles not just for users but also for
    groups (:func:`kotti.security.list_groups_callback`).
    """

    @reify
    def user(self) -> Optional[Principal]:
        """ Add the authenticated user to the request object.

        :result: the currently authenticated user
        :rtype: :class:`kotti.security.Principal` or whatever is returned by
                the custom principals database defined in the
                ``kotti.principals_factory`` setting
        """
        return get_user(self)

    def has_permission(
        self, permission: str, context: object = None
    ) -> Union[Allowed, Denied]:
        """ Check if the current request has the given permission on the
        current or explicitly passed context.  This is different from
        :meth:`pyramid.request.Request.has_permission`` in that a context other
        than the one bound to the request can be passed.  This allows to
        consider local roles for the check.

        :param permission: name of the permission to check
        :type permission: str

        :param context: context for which the permission is checked.
                        Defaults to the context on which the request invoked.
        :type context: :class:`kotti.resources.Node`

        :result: True if has_permission, False else
        :rtype: bool
        """

        from kotti.security import authz_context

        with authz_context(context, self):
            return super().has_permission(permission, context)



class BaseView:
    """ Very basic view class that can be subclassed.  Does nothing more than
    assignment of ``context`` and ``request`` to instance attributes on
    initialization. """

    def __init__(self, context, request):
        """ Constructor

        :param context: Context of the view
        :type context: :class:`kotti.resources.Node` or descendant for views on
                       content.

        :param request: Current request object
        :type request: :class:`kotti.request.Request`
        """

        self.context = context
        self.request = request


def includeme(config):
    """ Pyramid includeme hook.

    :param config: app config
    :type config: :class:`pyramid.config.Configurator`
    """

    config.add_static_view("static-kotti", "kotti:static")

    config.include("pyramid_deform")
    config.include("js.deform")

    config.include("kotti.views.util")



unsanitized = """
<h1>Title</h1>
<div class="teaser umlaut">Descrüptiön</div>
<p>
  Paragraph with
  <a href="internal.html" target="_blank">internal</a> and
  <a href="http://external.com/" target="_blank">external</a> links.
</p>
<marquee>Fancy marquee!</marquee>
<b SIZE=17 style="color: red">IMPORTANT!</b>
<script>
  alert('XSS!')
</script>
<p>Unclosed paragraph
"""


def _verify_no_html(sanitized):
    assert "<" not in sanitized
    assert "external links" in sanitized


def test_no_html():

    from kotti.sanitizers import no_html

    _verify_no_html(no_html(unsanitized))


def _verify_minimal_html(sanitized):

    from bleach_allowlist import all_tags
    from bleach_allowlist import markdown_tags
    from bleach_allowlist import print_tags

    for tag in set(all_tags) - set(markdown_tags) - set(print_tags):
        assert f"<{tag}" not in sanitized

    assert 'style=""' in sanitized
    assert '<a href="http://external.com/">' in sanitized
    assert "size" not in sanitized.lower()


def test_minmal_html():

    from kotti.sanitizers import minimal_html

    _verify_minimal_html(minimal_html(unsanitized))


def _verify_xss_protection(sanitized):

    assert "<script>" not in sanitized
    assert "<h1>Title</h1>" in sanitized
    assert '<a href="internal.html" target="_blank">internal</a>' in sanitized
    assert 'b size="17"' in sanitized
    assert "<p>Unclosed paragraph\n</p>" in sanitized


def test_xss_protection():

    from kotti.sanitizers import xss_protection

    _verify_xss_protection(xss_protection(unsanitized))


def test_default_config(unresolved_settings):

    assert "kotti.sanitizers" in unresolved_settings
    assert "kotti.sanitize_on_write" in unresolved_settings

    assert (
        unresolved_settings["kotti.sanitizers"]
        == "xss_protection:kotti.sanitizers.xss_protection minimal_html:kotti.sanitizers.minimal_html no_html:kotti.sanitizers.no_html"
    )  # noqa
    assert (
        unresolved_settings["kotti.sanitize_on_write"]
        == "kotti.resources.Document.body:xss_protection kotti.resources.Content.title:no_html kotti.resources.Content.description:no_html"
    )  # noqa


def test_setup_sanitizers(unresolved_settings):
    from kotti.sanitizers import _setup_sanitizers
    from kotti.sanitizers import minimal_html
    from kotti.sanitizers import no_html
    from kotti.sanitizers import xss_protection

    _setup_sanitizers(unresolved_settings)

    settings = unresolved_settings["kotti.sanitizers"]

    assert "minimal_html" in settings
    assert "no_html" in settings
    assert "xss_protection" in settings

    assert settings["minimal_html"] == minimal_html
    assert settings["no_html"] == no_html
    assert settings["xss_protection"] == xss_protection


def test_listeners(app, root, db_session):

    from kotti.resources import Document

    root["d"] = doc = Document(
        name="test", title="<h1>Title</h1>", description=unsanitized, body=unsanitized
    )

    db_session.flush()

    assert doc.title == "Title"
    _verify_no_html(doc.description)
    _verify_xss_protection(doc.body)

    # Test None title
    root["e"] = doc = Document(
        name="test", title=None, description=unsanitized, body=unsanitized
    )
    db_session.flush()
    assert doc.title == None


def test_sanitize(app, dummy_request):

    from kotti.sanitizers import sanitize
    from kotti.resources import Document
    from kotti.views.util import TemplateAPI

    _verify_no_html(sanitize(unsanitized, "no_html"))
    _verify_minimal_html(sanitize(unsanitized, "minimal_html"))
    _verify_xss_protection(sanitize(unsanitized, "xss_protection"))

    api = TemplateAPI(Document(), dummy_request)
    _verify_no_html(api.sanitize(unsanitized, "no_html"))
    _verify_minimal_html(api.sanitize(unsanitized, "minimal_html"))
    _verify_xss_protection(api.sanitize(unsanitized, "xss_protection"))



import pytest
from mock import patch
from webtest.forms import Upload

from kotti.testing import BASE_URL
from kotti.testing import user


class TestLogin:
    def test_it(self, app):
        res = app.post(
            "/@@login", dict(login="admin", password="secret", submit="submit")
        )
        assert res.status == "302 Found"
        res = res.maybe_follow()
        assert res.status == "200 OK"


class TestForbidden:
    def test_forbidden(self, app):
        app.get("/@@edit", headers={"Accept": "*/json"}, status=403)

    def test_forbidden_redirect(self, app):
        res = app.get("/@@edit", headers={"Accept": "text/html"}, status=302)
        assert res.location.startswith("http://localhost/@@login?came_from=")

    def test_forbidden_redirect_when_authenticated(self, app):
        with patch("pyramid.request.Request.authenticated_userid", "foo"):
            res = app.get("/@@edit", status=302)
        assert res.location == "http://localhost/@@forbidden"


# noinspection PyUnusedLocal
class TestUploadFile:
    def add_file(self, browser, contents=b"ABC"):
        file_ctrl = browser.getControl("File")
        file_ctrl.add_file(contents, "image/gif", "my_image.gif")
        browser.getControl("save").click()

    def add_file2(self, resp, contents=b"ABC"):
        form = resp.forms["deform"]
        form["upload"] = Upload("my_image.gif", contents, "image/gif")
        return form.submit().maybe_follow()

    @user("admin")
    def test_it(self, browser, filedepot):
        browser.open(BASE_URL + "/@@add_file")
        self.add_file(browser)
        assert "Item was added" in browser.contents

    @user("admin")
    def test_view_uploaded_file(self, browser, filedepot):
        browser.open(BASE_URL + "/@@add_file")
        self.add_file(browser)
        browser.getLink("View").click()
        browser.getLink("Download file").click()
        assert browser.contents == b"ABC"

    @user("admin")
    def test_tempstorage(self, app, filedepot, browser):
        # 1. upload a file
        browser.open(BASE_URL + "/@@add_file")
        self.add_file(browser, contents=b"DEF")

        # 2. edit the uploaded file, blanking out the title field
        browser.getLink("Edit").click()
        browser.getControl("Title").value = ""  # the error
        browser.getControl("save").click()
        assert "Your changes have been saved" not in browser.contents

        # 3. fill out the title field again
        browser.getControl("Title").value = "A title"
        browser.getControl("save").click()
        assert "Your changes have been saved" in browser.contents

        # 4. view / download the file
        browser.getLink("View").click()
        browser.getLink("Download file").click()
        assert browser.contents == b"DEF"

    @user("admin")
    def test_edit_uploaded_file(self, browser, filedepot):
        # 1. upload a file
        browser.open(BASE_URL + "/@@add_file")
        self.add_file(browser, contents=b"GHI")
        # 2. edit without making changes
        browser.getLink("Edit").click()
        browser.getControl("save").click()
        # 3. download
        browser.getLink("Download file").click()
        assert browser.contents == b"GHI"


class TestValidatorMaxLength:
    @user("admin")
    def test_title_max_length_document_ko(self, browser, filedepot):
        browser.open(BASE_URL + "/@@add_document")
        from kotti.resources import Node

        max_length = Node.name.property.columns[0].type.length
        browser.getControl("Title").value = "1" * (max_length + 1)  # the error
        browser.getControl("save").click()
        assert "Item was added" not in browser.contents

    @user("admin")
    def test_title_max_length_document_ok(self, browser):
        browser.open(BASE_URL + "/@@add_document")
        from kotti.resources import Node

        max_length = Node.name.property.columns[0].type.length
        browser.getControl("Title").value = "1" * max_length
        browser.getControl("save").click()
        assert "Item was added" in browser.contents

    @user("admin")
    def test_title_max_length_file_ok(self, browser):
        browser.open(BASE_URL + "/@@add_file")
        from kotti.resources import Node

        max_length = Node.name.property.columns[0].type.length
        browser.getControl("Title").value = "1" * max_length
        file_ctrl = browser.getControl("File")
        file_ctrl.add_file(b"abc", "image/gif", "my_image.gif")
        browser.getControl("save").click()
        assert "Item was added" in browser.contents

    @user("admin")
    def test_title_max_length_file_ko(self, browser):
        browser.open(BASE_URL + "/@@add_file")
        from kotti.resources import Node

        max_length = Node.name.property.columns[0].type.length
        browser.getControl("Title").value = "1" * (max_length + 1)  # the error
        file_ctrl = browser.getControl("File")
        file_ctrl.add_file(b"abc", "image/gif", "my_image.gif")
        browser.getControl("save").click()
        assert "Item was added" not in browser.contents


class TestBrowser:
    """ This is a one to one conversion of the former browsert.txt.
    These tests should definitively be rewritten and splitted into multiple
    methods to ease readability.
    """

    @staticmethod
    def _login(app, login, password):
        resp = app.get("/@@login")
        form = resp.forms["login-form"]
        form["login"] = login
        form["password"] = password
        resp = form.submit("submit")
        return resp

    @staticmethod
    def _select_children(resp, *child_idx):
        """ Mark the checkbox(es) of the rows in the ``contents`` view

        :param resp: response which's body contains the contents table
        :type resp: :class:`webtest.response.TestResponse`

        :param *child_idx: index of the child in the folder listing
                           (starting with 0)
        :type *child_idx: int

        :return: contents form with children selected (ready for submission)
        :rtype: :class:`webtest.forms.Form`
        """

        form = resp.forms["form-contents"]
        form["children"] = [
            cb._value for cb in [form.fields["children"][idx] for idx in child_idx]
        ]

        return form

    def test_login(self, webtest):

        # get the frontpage
        resp = webtest.app.get("/")
        assert "You have successfully installed Kotti." in resp.text

        # access a protected URL
        resp = webtest.app.get("/edit").maybe_follow()
        assert "Log in" in resp.text

        # submit the login form
        form = resp.forms["login-form"]
        assert form["came_from"].value == "http://localhost/edit"
        form["login"] = "admin"
        form["password"] = "secret"

        resp = form.submit("submit").maybe_follow()
        assert "Welcome, Administrator" in resp.text

        # log out
        resp = resp.click("Logout").maybe_follow()
        assert "You have been logged out" in resp.text

        # attempt to login with wrong credentials
        form = resp.forms["login-form"]
        assert form["came_from"].value == "http://localhost/edit"
        form["login"] = "admin"
        form["password"] = "funny"

        resp = form.submit("submit")
        assert "Welcome, Adminstrator" not in resp.text
        assert "Login failed" in resp.text

        # and again, this time with correct credentials
        form = resp.forms["login-form"]
        assert form["came_from"].value == "http://localhost/edit"
        form["login"] = "admin"
        form["password"] = "secret"
        resp = form.submit("submit").maybe_follow()
        assert "Welcome, Administrator" in resp.text

    @pytest.mark.user("admin")
    def test_exception_views(self, webtest):
        app = webtest.app
        resp = app.get("/this-does-not-exist", status=404)
        assert "<h1>Not Found</h1>" in resp.text

    @pytest.mark.user("admin")
    def test_content_management(self, webtest):

        from kotti.resources import Document
        from kotti.resources import File

        save_addable_document = Document.type_info.addable_to
        save_addable_file = File.type_info.addable_to

        app = webtest.app
        resp = app.get("/")

        # Add a document
        resp = resp.click("Document", index=0)
        assert "Add Document to Welcome to Kotti" in resp.text
        form = resp.forms["deform"]
        form["title"] = "Child One"
        resp = form.submit("save").maybe_follow()
        assert "Item was added" in resp.text
        assert resp.request.path == "/child-one/"

        # Edit the document
        resp = resp.click("Edit")
        assert "Edit Child One" in resp.text
        form = resp.forms["deform"]
        form["title"] = "First Child"
        resp = form.submit("save").maybe_follow()
        assert "Your changes have been saved" in resp.text

        resp = resp.click("Edit")
        form = resp.forms["deform"]
        assert form["title"].value == "First Child"
        resp = resp.click("Edit")
        assert "First Child" in resp.text

        # And now force a validation error:
        resp = resp.click("Edit")
        form = resp.forms["deform"]
        form["title"] = ""
        resp = form.submit("save")
        assert "There was a problem" in resp.text
        assert form["title"].value == ""

        # And now click the 'Cancel' button:
        resp = resp.click("Edit")
        form = resp.forms["deform"]
        form["title"] = "Firstborn"
        resp = form.submit("cancel").maybe_follow()
        assert "deform" not in resp.forms
        assert "Firstborn" not in resp.text

        # Now click the 'Cancel' button for an invalid form entry:
        resp = resp.click("Edit")
        form = resp.forms["deform"]
        form["title"] = ""
        resp = form.submit("cancel").maybe_follow()
        assert "deform" not in resp.forms
        assert "Firstborn" not in resp.text

        # Add two more documents, at different levels
        resp = app.get("/")
        resp = resp.click("Document", index=0)
        form = resp.forms["deform"]
        form["title"] = "Second Child"
        resp = form.submit("save").maybe_follow()
        assert "Item was added" in resp.text
        assert resp.request.path == "/second-child/"

        resp = resp.click("Document", index=0)
        form = resp.forms["deform"]
        form["title"] = "Grandchild"
        resp = form.submit("save").maybe_follow()
        assert "Item was added" in resp.text
        assert resp.request.path == "/second-child/grandchild/"

        # Add another grandchild with the same name:
        resp = app.get("/second-child/")
        resp = resp.click("Document", index=0)
        form = resp.forms["deform"]
        form["title"] = "Grandchild"
        resp = form.submit("save").maybe_follow()
        assert "Item was added" in resp.text
        assert resp.request.path == "/second-child/grandchild-1/"

        # There's no Add link if nothing can be added:
        resp = app.get("/second-child/grandchild-1/")
        try:
            Document.type_info.addable_to = ()
            File.type_info.addable_to = ()
            with pytest.raises(IndexError):
                resp.click(href="add_")
        finally:
            Document.type_info.addable_to = save_addable_document
            File.type_info.addable_to = save_addable_file

        # Add a file
        resp = app.get("/")
        resp = resp.click("File", index=0)
        form = resp.forms["deform"]
        form["description"] = "A tiny file"
        form["upload"] = Upload("tiny.txt", b"tiny", "text/plain")
        resp = form.submit("save").maybe_follow()
        assert "Item was added" in resp.text
        assert resp.request.path == "/tiny.txt/"

        # Add a file larger than maximum file size
        resp = app.get("/")
        resp = resp.click("File", index=0)
        form = resp.forms["deform"]
        form["title"] = "Huge file"
        form["description"] = "A huge file"
        form["upload"] = Upload("huge.txt", b"*" * (1024 * 1024 * 1024), "text/plain")
        resp = form.submit("save")
        assert "There was a problem" in resp.text
        assert "Maximum file size" in resp.text

        # Add tags to a document:
        resp = app.get("/second-child/")
        resp = resp.click("Document", index=0)
        form = resp.forms["deform"]
        form["title"] = "Grandchild"
        form["tags"] = ""
        form.submit("save").maybe_follow()
        resp = app.get("/second-child/")
        resp = resp.click("Document", index=0)
        form = resp.forms["deform"]
        form["title"] = "Grandchild"
        form["tags"] = "tag 1, tag 2,tag 3"
        form.submit("save").maybe_follow()
        resp = app.get("/second-child/grandchild-2/@@edit")
        form = resp.forms["deform"]
        assert "tag 1" in resp.text
        assert "tag 2" in resp.text
        assert "tag 3" in resp.text
        form["tags"] = "tag 1, tag 4, tag 5,tag 6, tag 7, übertag"
        form.submit("save").maybe_follow()
        resp = app.get("/second-child/grandchild-2/@@edit")
        assert 'value="tag 1,tag 4,tag 5,tag 6,tag 7,übertag"' in resp.text

        # Delete a document
        resp = app.get("/second-child/grandchild/")
        resp = resp.click("Delete", index=0)
        form = resp.forms["form-delete"]
        resp = form.submit("cancel")
        assert "Grandchild was deleted" not in resp.text
        resp = app.get("/second-child/grandchild/")
        resp = resp.click("Delete", index=0)
        form = resp.forms["form-delete"]
        resp = form.submit("delete", value="delete").maybe_follow()
        assert "Grandchild was deleted" in resp.text
        assert resp.request.path == "/second-child/"

    @pytest.mark.user("admin")
    def test_view_actions(self, webtest):

        from kotti.resources import Document
        from kotti.resources import File

        save_addable_document = Document.type_info.addable_to
        save_addable_file = File.type_info.addable_to

        app = webtest.app

        def _add_document(resp, title):
            resp = resp.click("Document", index=0)
            form = resp.forms["deform"]
            form["title"] = title
            resp = form.submit("save").maybe_follow()
            return resp

        # Add some documents

        resp = app.get("/")

        resp = _add_document(resp, "Child One")
        assert resp.request.path == "/child-one/"

        resp = _add_document(resp, "Child One One")
        assert resp.request.path == "/child-one/child-one-one/"

        resp = app.get("/child-one/")

        resp = _add_document(resp, "Child One Two")
        assert resp.request.path == "/child-one/child-one-two/"

        resp = app.get("/")

        resp = _add_document(resp, "Child Two")
        assert resp.request.path == "/child-two/"

        resp = _add_document(resp, "Child Two One")
        assert resp.request.path == "/child-two/child-two-one/"

        resp = app.get("/child-two/")

        resp = _add_document(resp, "Child Two Two")
        assert resp.request.path == "/child-two/child-two-two/"

        # On setup pages we can't use the actions:
        resp = app.get("/")
        resp = resp.click("User Management", index=0)
        with pytest.raises(IndexError):
            resp.click("Copy", index=0)
        resp = resp.click("Preferences", index=0)
        with pytest.raises(IndexError):
            resp.click("Copy", index=0)

        # Contents view actions

        # copy

        resp = app.get("/child-one")
        resp = resp.click("Contents", index=0)
        assert resp.request.path == "/child-one/@@contents"
        resp = resp.forms["form-contents"].submit("copy").maybe_follow()
        assert "You have to select items" in resp.text

        form = self._select_children(resp, 0)
        resp = form.submit("copy").maybe_follow()
        assert "Child One One was copied." in resp.text

        resp = app.get("/child-two/@@contents")
        assert "Child One One" not in resp.text
        resp = resp.forms["form-contents"].submit("paste").maybe_follow()
        assert "Child One One" in resp.text

        # cut

        resp = app.get("/child-two/@@contents")
        form = self._select_children(resp, 0)
        resp = form.submit("cut").maybe_follow()
        assert "cut." in resp.text

        resp = app.get("/child-one/@@contents")
        assert "Child Two One" not in resp.text

        resp = resp.forms["form-contents"].submit("paste").maybe_follow()
        assert "Child Two One" in resp.text

        with pytest.raises(IndexError):
            resp.click("Paste")

        resp = app.get("/child-one/@@contents")
        form = self._select_children(resp, 0, 1)
        form.submit("cut").maybe_follow()
        resp = app.get("/")
        resp = resp.click("Document", index=0)
        form = resp.forms["deform"]
        form["title"] = "Forth child"
        form.submit("save").maybe_follow()
        resp = app.get("/forth-child/@@contents")

        assert "Child One One" not in resp.text
        assert "Child One Two" not in resp.text

        resp.forms["form-contents"].submit("paste").maybe_follow()
        resp = app.get("/forth-child/@@contents")
        assert "Child One One" in resp.text
        assert "Child One Two" in resp.text

        resp = app.get("/child-one/@@contents")
        assert "Child One One" not in resp.text
        assert "Child One Two" not in resp.text

        # rename

        resp = app.get("/forth-child/@@contents")
        assert "child-one-one" in resp.text
        assert "Child One One" in resp.text
        assert "child-one-two" in resp.text
        assert "Child One Two" in resp.text
        form = self._select_children(resp, 0, 1)
        resp = form.submit("rename_nodes").maybe_follow()
        resp = resp.forms["form-rename-nodes"].submit("cancel").maybe_follow()
        assert "No changes were made." in resp.text
        assert resp.request.path == "/forth-child/@@contents"

        form = self._select_children(resp, 0, 1)
        resp = form.submit("rename_nodes").maybe_follow()
        form = resp.forms["form-rename-nodes"]
        form[form.submit_fields()[1][0]] = "child-one-three"
        form[form.submit_fields()[2][0]] = "Child One Three"
        form[form.submit_fields()[4][0]] = "child-one-four"
        form[form.submit_fields()[5][0]] = "Child One Four"
        resp = form.submit("rename_nodes").maybe_follow()
        assert resp.request.path == "/forth-child/@@contents"
        assert "child-one-one" not in resp.text
        assert "Child One One" not in resp.text
        assert "child-one-two" not in resp.text
        assert "Child One Two" not in resp.text
        assert "child-one-three" in resp.text
        assert "Child One Three" in resp.text
        assert "child-one-four" in resp.text
        assert "Child One Four" in resp.text

        resp = resp.click("File", index=0)
        form = resp.forms["deform"]
        form["description"] = "A file"
        form["upload"] = Upload("some.txt", b"something", "text/plain")
        form.submit("save").maybe_follow()
        resp = app.get("/forth-child/@@contents")
        form = self._select_children(resp, 0, 1, 2)
        resp = form.submit("delete_nodes").maybe_follow()
        resp = resp.forms["form-delete-nodes"].submit("cancel").maybe_follow()
        assert "No changes were made." in resp.text
        assert resp.request.path == "/forth-child/@@contents"
        form = self._select_children(resp, 0, 1, 2)
        resp = form.submit("delete_nodes").maybe_follow()
        assert "Are you sure" in resp.text
        resp = (
            resp.forms["form-delete-nodes"]
                .submit("delete_nodes", status=302)
                .maybe_follow()
        )
        assert "Child One Three was deleted." in resp.text
        assert "Child One Four was deleted." in resp.text
        assert "some.txt was deleted." in resp.text
        resp = app.get("/forth-child/@@contents")
        assert "Welcome to Kotti" in resp.text
        assert '<i class="glyphicon glyphicon-home"></i>' in resp.text
        assert '<i class="glyphicon glyphicon-folder-open"></i>' in resp.text
        assert (
            '<i class="glyphicon glyphicon-folder-close"></i>' not in resp.text
        )  # noqa

        # Navigation
        resp = resp.click("Navigate")
        resp = resp.click("Child Two", index=0)
        assert resp.request.path == "/child-two/"

    @pytest.mark.user("admin")
    def test_rename(self, webtest):

        app = webtest.app

        def _add_document(resp, title):
            resp = resp.click("Document", index=0)
            form = resp.forms["deform"]
            form["title"] = title
            resp = form.submit("save").maybe_follow()
            return resp

        # Add some documents

        resp = app.get("/")

        resp = _add_document(resp, "Child One")
        assert resp.request.path == "/child-one/"

        resp = _add_document(resp, "Child One One")
        assert resp.request.path == "/child-one/child-one-one/"

        # We can rename an item. Slashes will be stripped out.:
        resp = app.get("/child-one/child-one-one/")
        resp = resp.click("Rename", index=0)
        form = resp.forms["form-rename"]
        assert form["name"].value == "child-one-one"
        assert form["title"].value == "Child One One"
        form["name"] = "child/one/two"
        form["title"] = "Child One Two"
        resp = form.submit("rename").maybe_follow()
        assert "Item was renamed" in resp.text
        assert resp.request.path == "/child-one/childonetwo/"

        # We cannot rename the root:
        resp = app.get("/")
        with pytest.raises(IndexError):
            resp.click("Rename", index=0)

    @pytest.mark.user("admin")
    def test_clipboard(self, webtest):

        from kotti.resources import Document

        save_addable_document = Document.type_info.addable_to

        app = webtest.app

        def _add_document(resp, title):
            resp = resp.click("Document", index=0)
            form = resp.forms["deform"]
            form["title"] = title
            resp = form.submit("save").maybe_follow()
            return resp

        # Add some documents

        resp = app.get("/")

        resp = _add_document(resp, "Child One")
        assert resp.request.path == "/child-one/"

        resp = _add_document(resp, "Child One One")
        assert resp.request.path == "/child-one/child-one-one/"

        resp = app.get("/child-one/")

        resp = _add_document(resp, "Child One Two")
        assert resp.request.path == "/child-one/child-one-two/"

        resp = app.get("/")

        resp = _add_document(resp, "Child Two")
        assert resp.request.path == "/child-two/"

        resp = _add_document(resp, "Child Two One")
        assert resp.request.path == "/child-two/child-two-one/"

        resp = app.get("/child-two/")

        resp = _add_document(resp, "Child Two Two")
        assert resp.request.path == "/child-two/child-two-two/"

        # cut and paste
        resp = app.get("/child-two/child-two-two/")
        resp = resp.click("Cut", index=0).maybe_follow()
        assert "Child Two Two was cut" in resp.text

        resp = app.get("/child-one/")
        resp = resp.click("Paste", index=0).maybe_follow()
        assert "Child Two Two was pasted" in resp.text

        # the former location returns 404 now
        app.get("/child-two/child-two-two/", status=404)

        # copy and paste

        resp = app.get("/child-one/child-two-two/")
        resp = resp.click("Copy", index=0).maybe_follow()
        assert "Child Two Two was copied" in resp.text

        resp = app.get("/child-two/")
        resp = resp.click("Paste", index=0).maybe_follow()
        assert "Child Two Two was pasted" in resp.text

        # the former location exists again
        resp = app.get("/child-two/child-two-two/")
        assert resp.status_code == 200

        # paste again
        resp = app.get("/child-two/")
        resp = resp.click("Paste", index=0).maybe_follow()
        assert "Child Two Two was pasted" in resp.text

        # the pasted item was renamed to satisfy name uniqueness
        resp = app.get("/child-two/child-two-two-1/")
        assert resp.status_code == 200

        # We can also copy and paste items that contain children, like the whole site:
        resp = app.get("/")
        resp = resp.click("Copy", index=0).maybe_follow()
        assert "Welcome to Kotti was copied" in resp.text

        resp = app.get("/child-two/")
        resp = resp.click("Paste", index=0).maybe_follow()
        assert "Welcome to Kotti was pasted" in resp.text
        resp = app.get("/child-two/welcome-to-kotti/")
        assert resp.status_code == 200
        resp = app.get("/child-two/welcome-to-kotti/child-two/")
        assert resp.status_code == 200

        # Note how we can't cut and paste an item into itself:
        resp = app.get("/child-one/")
        resp = resp.click("Cut", index=0).maybe_follow()
        assert "Child One was cut" in resp.text
        with pytest.raises(IndexError):
            resp.click("Paste", index=0).maybe_follow()

        # And also not into one of its children
        resp = app.get("/child-one/child-one-one/")
        with pytest.raises(IndexError):
            resp.click("Paste", index=0).maybe_follow()

        # Whether we can paste or not also depends on the ``type_info.addable``
        # property:
        resp = app.get("/child-one/")
        resp.click("Copy", index=0).maybe_follow()
        resp = app.get("/child-one/child-one-two/")
        resp.click("Paste", index=0).maybe_follow()
        try:
            Document.type_info.addable_to = ()
            resp = app.get("/child-one/child-one-two/")
            with pytest.raises(IndexError):
                resp.click("Paste", index=0).maybe_follow()
        finally:
            Document.type_info.addable_to = save_addable_document

        # You can't cut the root of a site:
        resp = app.get("/child-one/")
        resp.click("Cut", index=0)
        resp = app.get("/")
        with pytest.raises(IndexError):
            resp.click("Cut", index=0)

    @pytest.mark.user("admin")
    def test_workflow_actions(self, webtest):

        app = webtest.app

        # Add a document
        resp = app.get("/")
        resp = resp.click("Document", index=0)
        form = resp.forms["deform"]
        form["title"] = "Child One"
        resp = form.submit("save").maybe_follow()
        assert "Item was added" in resp.text
        assert resp.request.path == "/child-one/"

        # Add more documents
        resp = app.get("/")
        resp = resp.click("Document", index=0)
        form = resp.forms["deform"]
        form["title"] = "Second Child"
        resp = form.submit("save").maybe_follow()
        assert resp.request.path == "/second-child/"

        resp = resp.click("Document", index=0)
        form = resp.forms["deform"]
        form["title"] = "Third Child"
        resp = form.submit("save").maybe_follow()
        assert resp.request.path == "/second-child/third-child/"

        resp = app.get("/second-child/third-child/")
        resp = resp.click("Document", index=0)
        form = resp.forms["deform"]
        form["title"] = "Grandchild 1"
        resp = form.submit("save").maybe_follow()
        assert resp.request.path == "/second-child/third-child/grandchild-1/"

        resp = app.get("/second-child/third-child/")
        resp = resp.click("Document", index=0)
        form = resp.forms["deform"]
        form["title"] = "Grandchild 2"
        resp = form.submit("save").maybe_follow()
        assert resp.request.path == "/second-child/third-child/grandchild-2/"

        resp = app.get("/second-child/third-child/")
        resp = resp.click("Document", index=0)
        form = resp.forms["deform"]
        form["title"] = "Grandchild 3"
        resp = form.submit("save").maybe_follow()
        assert resp.request.path == "/second-child/third-child/grandchild-3/"

        resp = app.get("/second-child/third-child/")
        resp = resp.click("Document", index=0)
        form = resp.forms["deform"]
        form["title"] = "Child One"
        resp = form.submit("save").maybe_follow()
        assert resp.request.path == "/second-child/third-child/child-one/"

        # Contents view change state actions (workflow)
        resp = resp.click("Second Child", index=0)
        resp = resp.click("Contents")
        assert (
            "/second-child/third-child/@@workflow-change?new_state=public" in resp.text
        )  # noqa

        resp = app.get(
            "/second-child/third-child/@@workflow-change?new_state=public"
        ).maybe_follow()  # noqa
        assert (
            "/second-child/third-child/@@workflow-change?new_state=private" in resp.text
        )  # noqa

        resp = app.get("/second-child/third-child/@@contents")
        form = self._select_children(resp, 0, 1, 2)
        resp = form.submit("change_state").maybe_follow()
        assert "Change workflow state" in resp.text

        resp = resp.forms["form-change-state"].submit("cancel").maybe_follow()
        assert "No changes were made." in resp.text
        assert resp.request.path == "/second-child/third-child/@@contents"

        form = self._select_children(resp, 0, 1, 2)
        resp = form.submit("change_state").maybe_follow()
        form = resp.forms["form-change-state"]
        form["children-to-change-state"] = []
        resp = form.submit("change_state").maybe_follow()
        assert "No changes were made." in resp.text

        form = self._select_children(resp, 0, 1, 2)
        resp = form.submit("change_state").maybe_follow()
        form = resp.forms["form-change-state"]
        form["to-state"] = "public"
        resp = form.submit("change_state").maybe_follow()
        assert "Your changes have been saved." in resp.text
        assert (
            "/second-child/third-child/grandchild-1/@@workflow-change?new_state=private"
            in resp.text
        )  # noqa
        assert (
            "/second-child/third-child/grandchild-2/@@workflow-change?new_state=private"
            in resp.text
        )  # noqa
        assert (
            "/second-child/third-child/grandchild-3/@@workflow-change?new_state=private"
            in resp.text
        )  # noqa

        resp = resp.click("Child One", index=1)
        assert (
            "/second-child/third-child/child-one/@@workflow-change?new_state=public"
            in resp.text
        )  # noqa

        app.get(
            "/second-child/third-child/child-one/@@workflow-change?new_state=public"
        )  # noqa
        resp = app.get("/second-child/third-child/@@contents")
        assert (
            "/second-child/third-child/child-one/@@workflow-change?new_state=private"
            in resp.text
        )  # noqa

        resp = resp.click("Child One", index=1)
        resp = resp.click("Document", index=0)
        form = resp.forms["deform"]
        form["title"] = "Sub child"
        resp = form.submit("save").maybe_follow()
        assert (
            "/second-child/third-child/child-one/sub-child/@@workflow-change?new_state=public"
            in resp.text
        )  # noqa

        resp = resp.click("Second Child", index=0)
        resp = resp.click("Contents")

        form = self._select_children(resp, 0)
        resp = form.submit("change_state").maybe_follow()
        form = resp.forms["form-change-state"]
        form["include-children"] = "include-children"
        form["to-state"] = "private"
        resp = form.submit("change_state").maybe_follow()
        assert "Your changes have been saved." in resp.text
        assert (
            "/second-child/third-child/@@workflow-change?new_state=public" in resp.text
        )  # noqa

        resp = app.get("/second-child/third-child/child-one/sub-child/")
        assert (
            "/second-child/third-child/child-one/sub-child/@@workflow-change?new_state=public"
            in resp.text
        )  # noqa

    def test_user_management(self, webtest, settings, dummy_mailer):
        from kotti import get_settings

        get_settings()["kotti.site_title"] = "Website des Kottbusser Tors"

        app = webtest.app
        resp = self._login(app, "admin", "secret").maybe_follow()

        # The user management screen is available through
        # the "Site Setup" submenu
        resp = resp.click("User Management")

        # Add Bob's Group and assign the ``viewer`` role
        form = resp.forms["deform_group_add"]
        form["name"] = "bobsgroup"
        form["title"] = "Bob's Group"
        form.get("checkbox", index=0).checked = True
        resp = form.submit("add_group").maybe_follow()
        assert "Bob's Group was added" in resp.text
        assert (
            resp.forms["form-global-roles"]["role::group:bobsgroup::role:viewer"].value
            == "on"
        )

        # And a Bob.
        # Only alphanumeric characters are allowed for the name
        form = resp.forms["deform_user_add"]
        form["name"] = "bob:"
        form["title"] = "Bob Dabolina"
        form["password"] = "secret"
        form["password-confirm"] = "secret"
        form["email"] = "bob@DABOLINA.com"
        resp = form.submit("add_user")
        assert "There was a problem" in resp.text

        # Use a valid username now.  Note how the checkbox to send a password
        # registration link is ticked:
        form = resp.forms["deform_user_add"]
        form["name"] = "Bob"
        form["email"] = "bob@DABOLINA.com"
        form["send_email"].checked = True
        resp = form.submit("add_user").maybe_follow()
        assert "Bob Dabolina was added" in resp.text

        # We cannot add Bob twice:
        form = resp.forms["deform_user_add"]
        form["name"] = "bob"
        form["title"] = "Bobby Brown"
        form["password"] = "secret"
        form["password-confirm"] = "secret"
        resp = form.submit("add_user")
        assert "A user with that name already exists" in resp.text

        # We cannot add Bob Dabolina's email twice:
        form = resp.forms["deform_user_add"]
        form["name"] = "bob2"
        form["title"] = "Bobby Brown"
        form["email"] = "bob@dabolina.com"
        form["password"] = "secret"
        form["password-confirm"] = "secret"
        resp = form.submit("add_user")
        assert "A user with that email already exists" in resp.text
        form = resp.forms["deform_user_add"]
        form["email"] = "bob@gmail.com"
        resp = form.submit("add_user").maybe_follow()
        assert "Bobby Brown was added" in resp.text

        # Searching for Bob will return both Bob and Bob's Group:
        form = resp.forms["form-principal-search"]
        form["query"] = "Bob"
        resp = form.submit("search")
        assert "Bob Dabolina" in resp.text
        assert "Bob's Group" in resp.text
        form = resp.forms["form-global-roles"]
        assert form["role::group:bobsgroup::role:viewer"].checked
        assert not form["role::group:bobsgroup::role:editor"].checked
        assert not form["role::bob::role:editor"].checked

        # We can click on the Bob's entry to edit the list of groups
        # he belongs to:
        resp = resp.click("Bob Dabolina")
        form = resp.forms["deform"]
        form["group"] = "bobsgroup"
        resp = form.submit("save").maybe_follow()
        assert "Your changes have been saved" in resp.text

        # We cannot update on the Bobby Brown's entry with duplicated email:
        resp = resp.click("Back to User Management")
        form = resp.forms["form-principal-search"]
        form["query"] = "Bobby"
        resp = form.submit("search")
        resp = resp.click("Bobby Brown")
        form = resp.forms["deform"]
        form["email"] = "bob@dabolina.com"
        resp = form.submit("save")
        assert "A user with that email already exists" in resp.text

        # If we cancel the edit of the user, we are redirected to the
        # user management screen:
        resp = resp.forms["deform"].submit("cancel").maybe_follow()
        assert "No changes were made." in resp.text
        assert resp.request.path == "/@@setup-users"

        # Back in user management we can see how Bob's inherited the
        # viewer role from Bob's Group:
        form = resp.forms["form-principal-search"]
        form["query"] = "Bob"
        resp = form.submit("search")
        form = resp.forms["form-global-roles"]
        assert form["role::group:bobsgroup::role:viewer"].checked
        assert "disabled" not in form["role::group:bobsgroup::role:viewer"].attrs
        assert form["role::bob::role:viewer"].checked
        assert "disabled" in form["role::bob::role:viewer"].attrs

        # We can click on the Bob's Group entry to edit an email address:
        resp = resp.click("Bob's Group")
        form = resp.forms["deform"]
        form["email"] = "bogsgroup@gmail.com"
        resp = form.submit("save").maybe_follow()
        assert "Your changes have been saved" in resp.text

        # Set password
        # ------------

        # Remember that we sent Bob an email for registration.
        # He can use it to set his own password:
        [email, email2] = dummy_mailer.outbox
        assert email.recipients == ['"Bob Dabolina" <bob@dabolina.com>']
        assert email.subject == "Your registration for Website des " "Kottbusser Tors"
        assert "Hello, Bob Dabolina!" in email.body
        assert "You are joining Website des Kottbusser Tors." in email.body
        assert "Click here to set your password and log in:" in email.body
        assert "@@set-password?" in email.body
        assert "?token=" in email.body or "&token=" in email.body

        # We'll use that link to set our password:
        resp.click("Logout").maybe_follow()
        path = email.body[email.body.index("http://localhost"):].split()[0][16:]
        resp = app.get(path)
        form = resp.forms["deform"]
        form["password"] = "newpassword"
        form["password-confirm"] = "newpasswoops"  # a silly error
        resp = form.submit("submit")
        assert "There was a problem with your submission" in resp.text
        form = resp.forms["deform"]
        form["password"] = "newpassword"
        form["password-confirm"] = "newpassword"
        resp = form.submit("submit").maybe_follow()
        assert "You have reset your password" in resp.text

        # We cannot use that link again:
        resp = app.get(path)
        form = resp.forms["deform"]
        form["password"] = "won't work"
        form["password-confirm"] = "won't work"
        resp = form.submit("submit")
        assert "Your password reset token may have expired" in resp.text

        # Log in as Bob with the new password:
        resp = self._login(app, "bOB", "newpassword").maybe_follow()
        assert "Welcome, Bob Dabolina" in resp.text
        app.get("/@@edit")
        resp = resp.click("Logout").maybe_follow()

        # The login form has a "Reset password" button.  Let's try it:
        resp = self._login(app, "bobby", "")
        assert "Login failed." in resp.text
        form = resp.forms["forgot-password-form"]
        form["login"] = "bob"
        resp = form.submit("reset-password").maybe_follow()
        assert "You should be receiving an email" in resp.text
        [email1, email2, email3] = dummy_mailer.outbox
        assert "Hello, Bob Dabolina!" in email3.body
        assert (
            "Click this link to reset your password at Website des "
            "Kottbusser Tors:" in email3.body
        )

        # User preferences
        # ----------------

        # The "Preferences" link leads us to a form where the user can change
        # their preferences so the user need to be authenticated:
        resp = self._login(app, "admin", "secret").maybe_follow()
        assert "Welcome, Administrator" in resp.text
        resp = app.get("/@@prefs")
        form = resp.forms["deform"]
        form["title"] = "Mr. Administrator"
        form["email"] = "admin@minad.com"
        resp = form.submit("save").maybe_follow()
        assert "Your changes have been saved" in resp.text

        # The email could not be used if already a users with that email exists:
        form = resp.forms["deform"]
        form["email"] = "bob@dabolina.com"
        resp = form.submit("save")

        # If the user cancel the process he will be redirected to the site root:
        form = resp.forms["deform"]
        resp = form.submit("cancel").maybe_follow()
        assert "Welcome to Kotti" in resp.text

        # Share
        # -----

        # The Share tab allows us to assign users and groups to roles:
        resp = app.get("/")
        resp = resp.click("Edit")
        resp = resp.click("Share")

        # We can search for users:
        form = resp.forms["form-share-search"]
        form["query"] = "Bob"
        resp = form.submit("search")

        # Bob and Bob's Group are listed now:
        assert "Bob Dabolina" in resp.text
        assert "Bob's Group" in resp.text

        # We add Bob's Group to Owners and Editors before taking away Owners
        # again:
        form = resp.forms["form-share-assign"]
        form["role::group:bobsgroup::role:owner"].checked = True
        form["role::group:bobsgroup::role:editor"].checked = True
        resp = form.submit("apply").maybe_follow()
        assert "Your changes have been saved" in resp.text
        resp = app.get(resp.request.path)  # >>> browser.reload()
        form = resp.forms["form-share-assign"]
        assert form["role::group:bobsgroup::role:owner"].checked
        assert form["role::group:bobsgroup::role:editor"].checked
        form["role::group:bobsgroup::role:owner"].checked = False
        resp = form.submit("apply").maybe_follow()
        assert "Your changes have been saved" in resp.text
        form = resp.forms["form-share-assign"]
        assert not form["role::group:bobsgroup::role:owner"].checked

        # Not making any changes will give us a different feedback message:
        resp = form.submit("apply")
        assert "Your changes have been saved" not in resp.text
        form = resp.forms["form-share-assign"]
        assert not form["role::group:bobsgroup::role:owner"].checked
        assert form["role::group:bobsgroup::role:editor"].checked

        # Bob should now have an inherited Editor role, because he's part of
        # Bob's Group:
        form = resp.forms["form-share-search"]
        form["query"] = "Bob Dabolina"
        resp = form.submit("search")
        form = resp.forms["form-share-assign"]
        assert form["role::bob::role:editor"].checked
        # TODO assert 'disabled' in form["role::bob::role:editor"].attrs
        # TODO assert not form["role::bob::role:owner"].checked

        # Lastly, let's take away the remaining Editor role from Bob's Group
        # again:
        assert "Bob's Group" in resp.text
        form["role::group:bobsgroup::role:editor"].checked = None
        resp = form.submit("apply").maybe_follow()
        assert "Your changes have been saved" in resp.text
        # TODO assert "Bob's Group" not in resp.text
        form = resp.forms["form-share-assign"]
        assert not form["role::group:bobsgroup::role:editor"].checked

        # Delete users
        # ------------

        resp = app.get("/")
        resp = resp.click("User Management")

        # We add a group and assign the ``Manager`` role to one:
        form = resp.forms["deform_group_add"]
        form["name"] = "pelayogroup"
        form["title"] = "Pelayo Group"
        form.get("checkbox", index=3).checked = True
        resp = form.submit("add_group").maybe_follow()
        assert "Pelayo Group was added" in resp.text
        form = resp.forms["form-global-roles"]
        assert form["role::group:pelayogroup::role:admin"].checked

        # And add some users to this group.
        form = resp.forms["deform_user_add"]
        form["name"] = "Bruce"
        form["title"] = "Bruce Pelayo"
        form["password"] = "secret"
        form["password-confirm"] = "secret"
        form["email"] = "bruce@pelayno.com"
        form["group"] = "pelayogroup"
        resp = form.submit("add_user").maybe_follow()
        assert "Bruce Pelayo was added" in resp.text

        form = resp.forms["deform_user_add"]
        form["name"] = "Brian"
        form["title"] = "Brian Pelayo"
        form["password"] = "secret"
        form["password-confirm"] = "secret"
        form["email"] = "brian@pelayno.com"
        form["group"] = "pelayogroup"
        resp = form.submit("add_user").maybe_follow()
        assert "Brian Pelayo was added" in resp.text

        # Lets login as Bruce and add some content:
        resp.click("Logout").maybe_follow()
        resp = self._login(app, "bruce", "secret").maybe_follow()
        assert "Welcome, Bruce Pelayo!" in resp.text
        resp = resp.click("Document", index=0)
        form = resp.forms["deform"]
        form["title"] = "Bruce one"
        form.submit("save").maybe_follow()
        resp = app.get("/")
        resp = resp.click("Document", index=0)
        form = resp.forms["deform"]
        form["title"] = "Bruce two"
        form.submit("save").maybe_follow()

        # Now let's delete Bruce::
        resp.click("Logout").maybe_follow()
        resp = self._login(app, "admin", "secret").maybe_follow()
        resp = resp.click("User Management")
        form = resp.forms["form-principal-search"]
        form["query"] = "Bruce"
        resp = form.submit("search").maybe_follow()
        resp = resp.click("Bruce Pelayo")
        resp = resp.forms["deform"].submit("delete").maybe_follow()
        assert "Delete <em>Bruce Pelayo</em>" in resp.text
        assert (
            "Are you sure you want to delete <span>User</span> "
            "<em>Bruce Pelayo</em>?" in resp.text
        )
        resp = resp.forms["form-delete-user"].submit("delete").maybe_follow()
        assert "User Bruce Pelayo was deleted." in resp.text
        form = resp.forms["form-principal-search"]
        form["query"] = "Bruce"
        resp = form.submit("search").maybe_follow()
        assert "No users or groups were found." in resp.text



import datetime
import time

import pytest
from mock import MagicMock
from mock import patch

from kotti.resources import File
from kotti.testing import Dummy
from kotti.testing import asset
from kotti.views.cache import set_max_age


def parse_expires(date_string):
    return datetime.datetime(
        *(time.strptime(date_string, "%a, %d %b %Y %H:%M:%S GMT")[0:6])
    )


def delta(date_string):
    now = datetime.datetime.utcnow()
    return parse_expires(date_string) - now


# noinspection PyUnusedLocal
@pytest.fixture
def cachetest_content(root, filedepot):
    image = asset("sendeschluss.jpg")
    root["textfile"] = File(b"file contents", "mytext.txt", "text/plain")
    root["image"] = File(image.read(), "sendeschluss.jpg", "image/jpeg")


class TestSetMaxAge:
    # noinspection PyUnresolvedReferences
    def test_preserve_existing_headers(self):
        response = Dummy(
            headers={"cache-control": "max-age=17,s-max-age=42,foo,bar=42"}
        )
        delta = datetime.timedelta(days=1)
        set_max_age(response, delta)

        assert "cache-control" in response.headers
        assert response.headers["cache-control"] == (
            "bar=42,foo,max-age=86400,s-max-age=42"
        )


class TestSetCacheHeaders:
    def test_caching_policy_chooser(self):
        from kotti.views.cache import set_cache_headers

        with patch("kotti.views.cache.caching_policy_chooser") as chooser:
            chooser.return_value = "Random policy"

            event = MagicMock()
            event.response.headers.get.return_value = None

            with pytest.raises(KeyError):
                set_cache_headers(event)

        chooser.assert_called_with(event.request.context, event.request, event.response)

    def test_caching_policy_chooser_raises(self):
        from kotti.views.cache import set_cache_headers

        def raiser(*args, **kw):
            raise Exception()

        with patch("kotti.views.cache.caching_policy_chooser") as chooser:
            chooser.side_effect = raiser

            event = MagicMock()
            event.response.headers.get.return_value = None

            with patch("kotti.views.cache.logger"):
                set_cache_headers(event)

        chooser.assert_called_with(event.request.context, event.request, event.response)

    def test_header_set_before(self):
        from kotti.views.cache import CACHE_POLICY_HEADER
        from kotti.views.cache import set_cache_headers

        event = MagicMock()
        event.response.headers = {CACHE_POLICY_HEADER: "Random policy"}

        with patch("kotti.views.cache.caching_policy_chooser") as chooser:
            with pytest.raises(KeyError):
                set_cache_headers(event)

        assert chooser.call_count == 0

    def test_request_has_no_context(self):
        from kotti.views.cache import set_cache_headers

        with patch("kotti.views.cache.caching_policy_chooser") as chooser:
            event = MagicMock()
            event.request = {}
            set_cache_headers(event)

            assert chooser.call_count == 0


class TestBrowser:
    def test_cache_unauth(self, webtest, cachetest_content):

        # html
        resp = webtest.app.get("/")
        assert resp.headers.get("X-Caching-Policy") == "Cache HTML"
        assert resp.headers.get("Cache-Control") == "max-age=0,s-maxage=3600"
        d = delta(resp.headers.get("Expires"))
        assert (d.days, d.seconds) < (0, 0)

        # resources
        resp = webtest.app.get("/static-kotti/base.css")
        assert resp.headers.get("X-Caching-Policy") == "Cache Resource"
        assert resp.headers.get("Cache-Control") == "max-age=2764800,public"
        d = delta(resp.headers.get("Expires"))
        assert (d.days, d.seconds) > (30, 0)
        assert "Last-Modified" in resp.headers

        # post
        resp = webtest.app.post("/", "")
        assert "X-Caching-Policy" not in resp.headers

        # 404
        resp = webtest.app.get("/this-isnt-here", status=404)
        assert "X-Caching-Policy" not in resp.headers

        # media content
        resp = webtest.app.get("/textfile/inline-view")
        assert resp.headers.get("X-Caching-Policy") == "Cache Media Content"
        assert resp.headers.get("Cache-Control") == "max-age=14400,public"
        d = delta(resp.headers.get("Expires"))
        assert (d.days, d.seconds) > (0, 14000)
        resp = webtest.app.get("/image/inline-view")
        assert resp.headers.get("X-Caching-Policy") == "Cache Media Content"
        assert resp.headers.get("Cache-Control") == "max-age=14400,public"
        d = delta(resp.headers.get("Expires"))
        assert (d.days, d.seconds) > (0, 14000)

    @pytest.mark.user("admin")
    def test_cache_auth(self, webtest, cachetest_content):

        # html
        resp = webtest.app.get("/")
        assert resp.headers.get("X-Caching-Policy") == "No Cache"

        # resources
        resp = webtest.app.get("/static-kotti/base.css")
        assert resp.headers.get("X-Caching-Policy") == "Cache Resource"

        # 404
        resp = webtest.app.get("/this-isnt-here", status=404)
        assert "X-Caching-Policy" not in resp.headers

        # media content
        resp = webtest.app.get("/textfile/inline-view")
        assert resp.headers.get("X-Caching-Policy") == "No Cache"
        assert resp.headers.get("Cache-Control") == "max-age=0,public"
        d = delta(resp.headers.get("Expires"))
        assert (d.days, d.seconds) <= (0, 0)
        resp = webtest.app.get("/image/inline-view")
        assert resp.headers.get("X-Caching-Policy") == "No Cache"
        assert resp.headers.get("Cache-Control") == "max-age=0,public"
        d = delta(resp.headers.get("Expires"))
        assert (d.days, d.seconds) <= (0, 0)



from mock import Mock
from mock import patch


class TestRequestCache:
    @property
    def cache_decorator(self):
        from kotti.util import request_cache

        return request_cache

    def test_it(self, dummy_request):
        from kotti.util import clear_cache

        called = []

        @self.cache_decorator(lambda a, b: (a, b))
        def my_fun(a, b):
            called.append((a, b))

        my_fun(1, 2)
        my_fun(1, 2)
        assert len(called) == 1
        my_fun(2, 1)
        assert len(called) == 2

        clear_cache()
        my_fun(1, 2)
        assert len(called) == 3

    def test_dont_cache(self, dummy_request):
        from kotti.util import DontCache

        called = []

        def dont_cache(a, b):
            raise DontCache

        @self.cache_decorator(dont_cache)
        def my_fun(a, b):
            called.append((a, b))

        my_fun(1, 2)
        my_fun(1, 2)
        assert len(called) == 2


class TestLRUCache(TestRequestCache):
    @property
    def cache_decorator(self):
        from kotti.util import lru_cache

        return lru_cache


class TestTitleToName:
    def setUp(self):
        from pyramid.threadlocal import get_current_registry
        from kotti.url_normalizer import url_normalizer

        r = get_current_registry()
        settings = r.settings = {}
        settings["kotti.url_normalizer"] = [url_normalizer]
        settings["kotti.url_normalizer.map_non_ascii_characters"] = False

    def test_max_length_40(self):
        self.setUp()
        from kotti.util import title_to_name

        assert len(title_to_name("a" * 50)) == 50

    def test_max_length_250(self):
        self.setUp()
        from kotti.util import title_to_name

        assert len(title_to_name("a" * 250)) == 240

    def test_max_length_255(self):
        self.setUp()
        from kotti.util import title_to_name

        assert len(title_to_name("a" * 255)) == 240

    def test_normal(self):
        self.setUp()
        from kotti.util import title_to_name

        assert title_to_name("Foo Bar") == "foo-bar"

    def test_max_length_40_no_default(self):
        self.setUp()
        from kotti.util import title_to_name

        assert len(title_to_name("a" * 50, max_length=40)) == 40

    def test_numbering(self):
        self.setUp()
        from kotti.util import title_to_name

        assert title_to_name("Report Part 1", blacklist=[]) == "report-part-1"
        assert (
            title_to_name("Report Part 1", blacklist=["report-part-1"])
            == "report-part-1-1"
        )
        assert (
            title_to_name("Report Part 3", blacklist=["report-part-3"])
            == "report-part-3-1"
        )
        assert (
            title_to_name(
                "Report Part 3", blacklist=["report-part-3", "report-part-3-1"]
            )
            == "report-part-3-2"
        )

    def test_disambiguate_name(self):
        from kotti.util import disambiguate_name

        assert disambiguate_name("foo") == "foo-1"
        assert disambiguate_name("foo-3") == "foo-4"


class TestCommand:
    def test_it(self):
        from kotti.util import command

        func = Mock()
        closer = Mock()
        with patch("kotti.util.docopt") as docopt:
            with patch("kotti.util.bootstrap") as bootstrap:
                docopt.return_value = {"<config_uri>": "app.ini"}
                bootstrap.return_value = {"closer": closer}
                assert command(func, "doc") == 0
                func.assert_called_with({"<config_uri>": "app.ini"})
                docopt.assert_called_with("doc")
                bootstrap.assert_called_with("app.ini")
                assert closer.call_count == 1


class TestTemplateStructure:
    def test_getattr(self):
        from kotti.util import TemplateStructure

        item = TemplateStructure("123")
        assert item.split("2") == ["1", "3"]


class TestLink:
    def test_link_selected(self):
        from kotti.util import Link
        from kotti.testing import DummyRequest

        req = DummyRequest()
        req.view_name = "manage"

        assert Link("manage").selected(Mock(__name__=None), req)

        req.view_name = "manage_cats"
        assert not Link("manage").selected(Mock(__name__=None), req)

        req.view_name = ""
        assert Link("").selected(Mock(__name__=None), req)

    def test_link_target(self):
        from kotti.util import Link

        assert Link("").target is None
        assert Link("", target="_blank").target == "_blank"



import pytest


def test_404_anon(webtest, root):

    resp = webtest.app.get("/non-existent", status=404)
    assert "Not Found" in resp.text


@pytest.mark.user("admin")
def test_404_admin(webtest, root):

    resp = webtest.app.get("/non-existent", status=404)
    assert "Not Found" in resp.text


def test_404_api_root(db_session, dummy_request):
    from kotti.resources import get_root
    from kotti.views.util import TemplateAPI
    from pyramid.httpexceptions import HTTPNotFound

    api = TemplateAPI(HTTPNotFound(), dummy_request)

    assert api.root == get_root()



from pyramid.exceptions import Forbidden
from pyramid.httpexceptions import HTTPBadRequest
from pytest import raises
from webob.multidict import MultiDict

from kotti.testing import DummyRequest


class TestAddableTypes:
    def test_view_permitted_yes(self, config, root):
        from kotti.resources import Document

        config.testing_securitypolicy(permissive=True)
        config.include("kotti.views.edit.content")
        request = DummyRequest()
        assert Document.type_info.addable(root, request) is True

    def test_view_permitted_no(self, config, root):
        from kotti.resources import Document

        config.testing_securitypolicy(permissive=False)
        config.include("kotti.views.edit.content")
        request = DummyRequest()
        assert Document.type_info.addable(root, request) is False

    def test_addable_views_registered_to_some_context(self, config, root):

        from kotti.resources import Document, File

        _saved = Document.type_info.addable_to
        Document.type_info.addable_to = ["File"]

        config.testing_securitypolicy(permissive=True)
        config.add_view(
            lambda context, request: {},
            name=Document.type_info.add_view,
            permission="add",
            renderer="kotti:templates/edit/node.pt",
            context=File,
        )
        request = DummyRequest()

        assert Document.type_info.addable(Document(), request) is False
        assert Document.type_info.addable(File(), request) is True

        Document.type_info.addable_to = _saved


class TestNodePaste:
    def test_get_non_existing_paste_item(self, root):
        from kotti.util import get_paste_items

        request = DummyRequest()
        request.session["kotti.paste"] = ([1701], "copy")
        item = get_paste_items(root, request)
        assert item == []

    def test_paste_non_existing_node(self, root):
        from kotti.views.edit.actions import NodeActions

        request = DummyRequest()

        for index, action in enumerate(["copy", "cut"]):
            request.session["kotti.paste"] = ([1701], "copy")
            response = NodeActions(root, request).paste_nodes()
            assert response.status == "302 Found"
            assert len(request.session["_f_error"]) == index + 1

    def test_paste_without_edit_permission(self, config, root):
        from kotti.views.edit.actions import NodeActions

        request = DummyRequest()
        request.params["paste"] = "on"
        config.testing_securitypolicy(permissive=False)

        # We need to have the 'edit' permission on the original object
        # to be able to cut and paste:
        request.session["kotti.paste"] = ([1], "cut")
        view = NodeActions(root, request)
        with raises(Forbidden):
            view.paste_nodes()

        # We don't need 'edit' permission if we're just copying:
        request.session["kotti.paste"] = ([1], "copy")
        response = NodeActions(root, request).paste_nodes()
        assert response.status == "302 Found"


class TestNodeRename:
    def setUp(self):
        from pyramid.threadlocal import get_current_registry
        from kotti.url_normalizer import url_normalizer

        r = get_current_registry()
        settings = r.settings = {}
        settings["kotti.url_normalizer"] = [url_normalizer]
        settings["kotti.url_normalizer.map_non_ascii_characters"] = False

    def test_rename_to_empty_name(self, root):
        from kotti.resources import Document
        from kotti.views.edit.actions import NodeActions

        child = root["child"] = Document(title="Child")
        request = DummyRequest()
        request.params["rename"] = "on"
        request.params["name"] = ""
        request.params["title"] = "foo"
        NodeActions(child, request).rename_node()
        assert request.session.pop_flash("error") == ["Name and title are required."]

    def test_multi_rename(self, root):
        from kotti.resources import Document
        from kotti.views.edit.actions import NodeActions

        self.setUp()
        root["child1"] = Document(title="Child 1")
        root["child2"] = Document(title="Child 2")
        request = DummyRequest()
        request.POST = MultiDict()
        id1 = str(root["child1"].id)
        id2 = str(root["child2"].id)
        request.POST.add("children-to-rename", id1)
        request.POST.add("children-to-rename", id2)
        request.POST.add(id1 + "-name", "")
        request.POST.add(id1 + "-title", "Unhappy Child")
        request.POST.add(id2 + "-name", "happy-child")
        request.POST.add(id2 + "-title", "")
        request.POST.add("rename_nodes", "rename_nodes")
        NodeActions(root, request).rename_nodes()
        assert request.session.pop_flash("error") == ["Name and title are required."]

        request.POST.add(id1 + "-name", "unhappy-child")
        request.POST.add(id1 + "-title", "Unhappy Child")
        request.POST.add(id2 + "-name", "happy-child")
        request.POST.add(id2 + "-title", "Happy Child")
        request.POST.add("rename_nodes", "rename_nodes")
        NodeActions(root, request).rename_nodes()
        assert request.session.pop_flash("success") == ["Your changes have been saved."]


class TestNodeDelete:
    def test_multi_delete(self, root):
        from kotti.resources import Document
        from kotti.resources import File
        from kotti.views.edit.actions import NodeActions

        root["child1"] = Document(title="Child 1")
        root["child2"] = Document(title="Child 2")
        root["file1"] = File(title="File 1")

        request = DummyRequest()
        request.POST = MultiDict()
        id1 = str(root["child1"].id)
        id2 = str(root["child2"].id)
        id3 = str(root["file1"].id)
        request.POST.add("delete_nodes", "delete_nodes")
        NodeActions(root, request).delete_nodes()
        assert request.session.pop_flash("info") == ["Nothing was deleted."]

        request.POST.add("children-to-delete", id1)
        request.POST.add("children-to-delete", id2)
        request.POST.add("children-to-delete", id3)
        NodeActions(root, request).delete_nodes()
        assert request.session.pop_flash("success") == [
            "${title} was deleted.",
            "${title} was deleted.",
            "${title} was deleted.",
        ]


class TestNodeMove:
    def test_move_up(self, root):
        from kotti.resources import Document
        from kotti.views.edit.actions import NodeActions

        root["child1"] = Document(title="Child 1")
        root["child2"] = Document(title="Child 2")
        assert root["child1"].position < root["child2"].position

        request = DummyRequest()
        request.session["kotti.selected-children"] = [str(root["child2"].id)]
        NodeActions(root, request).up()
        assert request.session.pop_flash("success") == ["${title} was moved."]
        assert root["child1"].position > root["child2"].position

    def test_move_down(self, root):
        from kotti.resources import Document
        from kotti.views.edit.actions import NodeActions

        root["child1"] = Document(title="Child 1")
        root["child2"] = Document(title="Child 2")
        root["child3"] = Document(title="Child 3")
        assert root["child1"].position < root["child3"].position
        assert root["child2"].position < root["child3"].position

        request = DummyRequest()
        ids = [str(root["child1"].id), str(root["child2"].id)]
        request.session["kotti.selected-children"] = ids
        NodeActions(root, request).down()
        assert request.session.pop_flash("success") == [
            "${title} was moved.",
            "${title} was moved.",
        ]
        assert root["child1"].position > root["child3"].position
        assert root["child2"].position > root["child3"].position

    def test_move_child_position_post(self, root, db_session):

        import transaction

        from kotti.resources import Document
        from kotti.resources import get_root
        from kotti.views.edit.actions import move_child_position

        # Create some documents
        root["child1"] = Document(title="Child 1")
        root["child2"] = Document(title="Child 2")
        root["child3"] = Document(title="Child 3")
        root["child4"] = Document(title="Child 4")
        root["child5"] = Document(title="Child 5")

        assert [c.position for c in root._children] == [0, 1, 2, 3, 4]
        assert [c.name for c in root._children] == [
            "child1",
            "child2",
            "child3",
            "child4",
            "child5",
        ]

        request = DummyRequest()

        # Move down
        request.POST = {"from": "0", "to": "3"}
        result = move_child_position(root, request)
        transaction.commit()
        root = get_root()
        assert result["result"] == "success"
        assert [c.position for c in root._children] == [0, 1, 2, 3, 4]
        assert [c.name for c in root._children] == [
            "child2",
            "child3",
            "child4",
            "child1",
            "child5",
        ]

        # Move up
        request.POST = {"from": "4", "to": "0"}
        move_child_position(root, request)
        transaction.commit()
        root = get_root()
        assert result["result"] == "success"
        assert [c.position for c in root._children] == [0, 1, 2, 3, 4]
        assert [c.name for c in root._children] == [
            "child5",
            "child2",
            "child3",
            "child4",
            "child1",
        ]

        # Invalid param value
        request.POST = {"from": "a", "to": "3"}
        result = move_child_position(root, request)
        assert result["result"] == "error"

        request.POST = {"from": "0", "to": "10"}
        result = move_child_position(root, request)
        assert result["result"] == "error"

        request.POST = {"from": "10", "to": "0"}
        result = move_child_position(root, request)
        assert result["result"] == "error"

        # Missing param
        request.POST = {"from": "a"}
        result = move_child_position(root, request)
        assert result["result"] == "error"

        # we have to clean up, because we committed transactions
        del root["child1"]
        del root["child2"]
        del root["child3"]
        del root["child4"]
        del root["child5"]
        transaction.commit()

    def test_move_child_position_json(self, root, db_session):

        import transaction

        from kotti.resources import Document
        from kotti.resources import get_root
        from kotti.views.edit.actions import move_child_position

        # Create some documents
        root["child1"] = Document(title="Child 1")
        root["child2"] = Document(title="Child 2")
        root["child3"] = Document(title="Child 3")
        root["child4"] = Document(title="Child 4")
        root["child5"] = Document(title="Child 5")

        assert [c.position for c in root._children] == [0, 1, 2, 3, 4]
        assert [c.name for c in root._children] == [
            "child1",
            "child2",
            "child3",
            "child4",
            "child5",
        ]

        request = DummyRequest()

        # Move down
        request.json_body = {"from": "0", "to": "3"}
        result = move_child_position(root, request)
        transaction.commit()
        root = get_root()
        assert result["result"] == "success"
        assert [c.position for c in root._children] == [0, 1, 2, 3, 4]
        assert [c.name for c in root._children] == [
            "child2",
            "child3",
            "child4",
            "child1",
            "child5",
        ]

        # Move up
        request.json_body = {"from": "4", "to": "0"}
        move_child_position(root, request)
        transaction.commit()
        root = get_root()
        assert result["result"] == "success"
        assert [c.position for c in root._children] == [0, 1, 2, 3, 4]
        assert [c.name for c in root._children] == [
            "child5",
            "child2",
            "child3",
            "child4",
            "child1",
        ]

        # Invalid param value
        request.json_body = {"from": "a", "to": "3"}
        result = move_child_position(root, request)
        assert result["result"] == "error"

        request.json_body = {"from": "0", "to": "10"}
        result = move_child_position(root, request)
        assert result["result"] == "error"

        request.json_body = {"from": "10", "to": "0"}
        result = move_child_position(root, request)
        assert result["result"] == "error"

        # Missing param
        request.json_body = {"from": "a"}
        result = move_child_position(root, request)
        assert result["result"] == "error"

        # we have to clean up, because we committed transactions
        del root["child1"]
        del root["child2"]
        del root["child3"]
        del root["child4"]
        del root["child5"]
        transaction.commit()


class TestNodeShowHide:
    def test_show_hide(self, root):
        from kotti.resources import Document
        from kotti.views.edit.actions import NodeActions

        root["child1"] = Document(title="Child 1")
        assert root["child1"].in_navigation is True

        request = DummyRequest()
        request.session["kotti.selected-children"] = [str(root["child1"].id)]
        NodeActions(root, request).hide()
        assert request.session.pop_flash("success") == [
            "${title} is no longer visible in the navigation."
        ]
        assert root["child1"].in_navigation is False

        request.session["kotti.selected-children"] = [str(root["child1"].id)]
        NodeActions(root, request).show()
        assert request.session.pop_flash("success") == [
            "${title} is now visible in the navigation."
        ]
        assert root["child1"].in_navigation is True


class TestNodeShare:
    def test_roles(self, root):
        from kotti.views.users import share_node
        from kotti.security import SHARING_ROLES

        # The 'share_node' view will return a list of available roles
        # as defined in 'kotti.security.SHARING_ROLES'
        request = DummyRequest()
        assert [
            r.name for r in share_node(root, request)["available_roles"]
        ] == SHARING_ROLES

    def test_search(self, extra_principals, root, db_session):
        from kotti.security import get_principals
        from kotti.security import set_groups
        from kotti.testing import DummyRequest
        from kotti.views.users import share_node

        request = DummyRequest()
        P = get_principals()

        # Search for "Bob", which will return both the user and the
        # group, both of which have no roles:
        request.params["search"] = ""
        request.params["query"] = "Bob"
        entries = share_node(root, request)["entries"]
        assert len(entries) == 2
        assert entries[0][0] == P["bob"]
        assert entries[0][1] == ([], [])
        assert entries[1][0] == P["group:bobsgroup"]
        assert entries[1][1] == ([], [])

        # We make Bob an Editor in this context, and Bob's Group
        # becomes global Admin:
        set_groups("bob", root, ["role:editor"])
        db_session.flush()
        P["group:bobsgroup"].groups = ["role:admin"]
        db_session.flush()
        entries = share_node(root, request)["entries"]
        assert len(entries) == 2
        assert entries[0][0] == P["bob"]
        assert entries[0][1] == (["role:editor"], [])
        assert entries[1][0] == P["group:bobsgroup"]
        assert entries[1][1] == (["role:admin"], ["role:admin"])

        # A search that doesn't return any items will still include
        # entries with existing local roles:
        request.params["query"] = "Weeee"
        entries = share_node(root, request)["entries"]
        assert len(entries) == 1
        assert entries[0][0] == P["bob"]
        assert entries[0][1] == (["role:editor"], [])
        assert request.session.pop_flash("info") == ["No users or groups were found."]

        # It does not, however, include entries that have local group
        # assignments only:
        set_groups("frank", root, ["group:franksgroup"])
        request.params["query"] = "Weeee"
        entries = share_node(root, request)["entries"]
        assert len(entries) == 1
        assert entries[0][0] == P["bob"]

    def test_apply(self, extra_principals, root):
        from kotti.security import list_groups
        from kotti.security import set_groups
        from kotti.views.users import share_node

        request = DummyRequest()

        request.params["apply"] = ""
        request.params["csrf_token"] = request.session.get_csrf_token()
        share_node(root, request)
        assert request.session.pop_flash("info") == ["No changes were made."]
        assert list_groups("bob", root) == []
        set_groups("bob", root, ["role:special"])

        request.params["role::bob::role:owner"] = "1"
        request.params["role::bob::role:editor"] = "1"
        request.params["orig-role::bob::role:owner"] = ""
        request.params["orig-role::bob::role:editor"] = ""

        share_node(root, request)
        assert request.session.pop_flash("success") == ["Your changes have been saved."]
        assert set(list_groups("bob", root)) == {
            "role:owner",
            "role:editor",
            "role:special",
        }

        # We cannot set a role that's not displayed, even if we forged
        # the request:
        request.params["role::bob::role:admin"] = "1"
        request.params["orig-role::bob::role:admin"] = ""
        with raises(Forbidden):
            share_node(root, request)
        assert set(list_groups("bob", root)) == {
            "role:owner",
            "role:editor",
            "role:special",
        }

    def test_csrf(self, extra_principals, root, dummy_request):
        """ Test if a CSRF token is present and checked on submission """
        from kotti.views.users import share_node

        result = share_node(root, dummy_request)
        assert "csrf_token" in result

        dummy_request.params["apply"] = ""

        with raises(HTTPBadRequest):
            share_node(root, dummy_request)

        dummy_request.params[
            "csrf_token"
        ] = dummy_request.session.get_csrf_token()  # noqa
        result = share_node(root, dummy_request)
        assert result["csrf_token"] == dummy_request.session.get_csrf_token()



from typing import List
from typing import Optional
from typing import Union

from fanstatic import Group
from fanstatic import Library
from fanstatic import Resource
from js.angular import angular
from js.bootstrap import bootstrap_css
from js.bootstrap import bootstrap_js
from js.fineuploader import fineuploader
from js.html5shiv import html5shiv
from js.jquery import jquery
from js.jquery_form import jquery_form
from js.jquery_tablednd import jquery_tablednd
from js.jqueryui import bootstrap as jqueryui_bootstrap_theme
from js.jqueryui_tagit import tagit as ui_tagit

# This is needed until ``kotti.views.form.deferred_tag_it_widget`` is converted
# to a class with a ``requirements`` attribute (that would be auto_needed by
# ``js.deform[_bootstrap]``).
tagit = Group([ui_tagit, jqueryui_bootstrap_theme])

# Kotti's resources
lib_kotti = Library("kotti", "static")
contents_view_js = Resource(
    lib_kotti,
    "contents.js",
    depends=[jquery_tablednd],
    minified="contents.min.js",
    bottom=True,
)
base_css = Resource(
    lib_kotti,
    "base.css",
    depends=[bootstrap_css],
    minified="base.min.css",
    dont_bundle=True,
)
edit_css = Resource(lib_kotti, "edit.css", depends=[base_css], minified="edit.min.css")
view_css = Resource(lib_kotti, "view.css", depends=[base_css], minified="view.min.css")

# Resources for content upload views
upload_js = Resource(
    lib_kotti,
    "upload.js",
    depends=[angular, fineuploader],
    # minified="upload.min.js", needs special minifying
    bottom=True,
)
upload_css = Resource(
    lib_kotti, "upload.css", depends=[base_css], minified="upload.min.css"
)
upload = Group([upload_js, upload_css])


class NeededGroup:
    """A collection of fanstatic resources that supports
       dynamic appending of resources after initialization"""

    def __init__(
        self, resources: Optional[List[Union[Resource, "NeededGroup"]]] = None
    ):  # noqa

        if resources is None:
            resources = []

        if not isinstance(resources, list):
            raise ValueError(
                "resources must be a list of fanstatic.Resource "
                "and/or fanstatic.Group objects"
            )

        self.resources = []

        for resource in resources:
            self.add(resource)

    def add(self, resource: Union["NeededGroup", Resource]):
        """resource may be a:

            - :class:`fanstatic.Resource` object or
            - :class:`fanstatic.Group` object"""

        if isinstance(resource, self.__class__):
            self.resources = self.resources + resource.resources
        elif isinstance(resource, (Resource, Group)):
            self.resources.append(resource)
        else:
            raise ValueError(
                "resource must be a NeededGroup,"
                "fanstatic.Resource or fanstatic.Group object"
            )

    def need(self):  # pragma: no cover
        # this is tested in fanstatic itself; we should add browser tests
        # for `view_needed` and `edit_needed` (see below)
        Group(self.resources).need()


view_needed_css = NeededGroup([view_css])
view_needed_js = NeededGroup([jquery, bootstrap_js, html5shiv])
view_needed = NeededGroup([view_needed_css, view_needed_js])

edit_needed_css = NeededGroup([edit_css, jqueryui_bootstrap_theme])
edit_needed_js = NeededGroup([jquery, bootstrap_js, html5shiv, jquery_form])
edit_needed = NeededGroup([edit_needed_css, edit_needed_js])



import colander
from mock import Mock

from kotti.testing import DummyRequest


class DummyContext:
    view_name = "view_name"
    tags = ["tag 1", "tag 2", "tag 3"]


class TestTags:
    def test_empty(self, root):
        assert root.tags == []

    def test_tags(self, db_session):
        from kotti.resources import Tag

        new_tag = Tag(title="test tag")
        assert str(new_tag) == "<Tag ('test tag')>"

    def test_add(self, root):
        from kotti.resources import Tag, TagsToContents

        root.tags = ["tag 1", "tag 2"]
        result = Tag.query.filter(TagsToContents.item == root).all()
        assert result[0].items == [root]
        assert root.tags == ["tag 1", "tag 2"]
        assert len(Tag.query.all()) == 2

    def test_edit(self, root):
        root.tags = ["tag 1", "tag_2"]
        assert root._tags[0].tag.title == "tag 1"
        root.tags = ["edited tag", "tag_2"]
        assert root._tags[0].tag.title == "edited tag"

    def test_association_proxy(self, root):
        from kotti.resources import Tag, TagsToContents, Content

        root["content_1"] = Content()
        root["content_1"].tags = ["tag 1", "tag 2"]
        assert root["content_1"].tags == ["tag 1", "tag 2"]
        assert type(root["content_1"]._tags[0]) == TagsToContents
        assert type(root["content_1"]._tags[0].tag) == Tag
        assert root["content_1"]._tags[0].tag.title == "tag 1"
        assert root["content_1"]._tags[0].position == 0
        assert root["content_1"]._tags[1].tag.title == "tag 2"
        assert root["content_1"]._tags[1].position == 1
        assert len(root["content_1"]._tags) == 2

        root["content_2"] = Content()
        root["content_2"].tags = ["tag 1", "tag 3"]
        assert len(root["content_2"]._tags) == 2
        assert root["content_2"]._tags[0].tag.title == "tag 1"
        assert root["content_2"]._tags[0].position == 0
        assert root["content_2"]._tags[1].tag.title == "tag 3"
        assert root["content_2"]._tags[1].position == 1
        assert len(Tag.query.all()) == 3

    def test_delete_tag_doesnt_touch_content(self, root, db_session):
        from kotti.resources import Tag, Content

        root["content_1"] = Content()
        root["content_1"].tags = ["my tag"]

        assert Content.query.filter_by(name="content_1").count() == 1
        db_session.delete(Tag.query.filter_by(title="my tag").one())
        assert Content.query.filter_by(name="content_1").count() == 1

    def test_delete_content_deletes_orphaned_tags(self, root, events):
        from kotti.resources import Tag, Content

        root["content_1"] = Content()
        root["content_2"] = Content()
        root["content_1"].tags = ["tag 1", "tag 2"]
        root["content_2"].tags = ["tag 2"]
        assert Tag.query.count() == 2
        del root["content_1"]
        assert Tag.query.one().title == "tag 2"

    def test_delete_tag_assignment_doesnt_touch_content(self, root, db_session):
        from kotti.resources import Tag, TagsToContents, Content

        root["content_1"] = Content()
        root["content_1"].tags = ["my tag"]

        assert Tag.query.count() == 1
        assert Content.query.filter_by(name="content_1").count() == 1
        db_session.delete(TagsToContents.query.one())
        assert Content.query.filter_by(name="content_1").count() == 1

    def test_delete_tag_assignment_delete_tag(self, root, events, db_session):
        from kotti.resources import Tag, TagsToContents, Content

        root["content_1"] = Content()
        root["content_1"].tags = ["my tag"]

        assert Tag.query.count() == 1
        db_session.delete(TagsToContents.query.one())
        assert Tag.query.count() == 0

    def test_copy_content_copy_tags(self, root, db_session):
        from kotti.resources import Tag, TagsToContents, Content

        root["content_1"] = Content()
        root["content_1"].tags = ["my tag"]
        assert Tag.query.count() == 1
        assert TagsToContents.query.count() == 1

        root["content_2"] = root["content_1"].copy()
        db_session.flush()
        assert root["content_1"].tags == ["my tag"]
        assert root["content_2"].tags == ["my tag"]
        assert Tag.query.count() == 1
        assert TagsToContents.query.count() == 2

    def test_cut_and_paste_content_copy_tags(self, root):
        from kotti.resources import Tag, TagsToContents, Content
        from kotti.views.edit.actions import NodeActions

        root["folder_1"] = Content()
        root["content_1"] = Content()
        root["content_1"].tags = ["my tag"]
        assert Tag.query.count() == 1
        assert TagsToContents.query.count() == 1

        request = DummyRequest()
        request.params["paste"] = "on"
        request.session["kotti.paste"] = ([root["content_1"].id], "cut")
        NodeActions(root["folder_1"], request).paste_nodes()
        assert root["folder_1"]["content_1"].tags == ["my tag"]
        assert Tag.query.count() == 1
        assert TagsToContents.query.count() == 1

    def test_copy_and_paste_content_copy_tags(self, root, events):
        from kotti.resources import Tag, TagsToContents, Content
        from kotti.views.edit.actions import NodeActions

        root["folder_1"] = Content()
        root["content_1"] = Content()
        root["content_1"].tags = ["my tag"]
        assert Tag.query.count() == 1
        assert TagsToContents.query.count() == 1

        request = DummyRequest()
        request.params["paste"] = "on"
        request.session["kotti.paste"] = ([root["content_1"].id], "copy")
        NodeActions(root["folder_1"], request).paste_nodes()
        assert root["content_1"].tags == ["my tag"]
        assert root["folder_1"]["content_1"].tags == ["my tag"]
        assert Tag.query.count() == 1
        assert TagsToContents.query.count() == 2

    def test_delete_content_delete_tags_and_assignments(self, root, events):
        from kotti.resources import Tag, TagsToContents, Content
        from kotti.views.edit.actions import NodeActions

        root["folder_1"] = Content()
        root["folder_1"].tags = ["first tag"]
        root["folder_1"]["content_1"] = Content()
        root["folder_1"]["content_1"].tags = ["second tag"]
        root["folder_1"]["content_2"] = Content()
        root["folder_1"]["content_2"].tags = ["third tag"]
        assert Tag.query.count() == 3
        assert TagsToContents.query.count() == 3

        request = DummyRequest()
        request.POST["delete"] = "delete"
        NodeActions(root["folder_1"], request).delete_node()
        assert Tag.query.count() == 0
        assert TagsToContents.query.count() == 0

    def test_get_content_items_from_tag(self, root):
        from kotti.resources import Tag, Content

        root["folder_1"] = Content()
        root["folder_1"].tags = ["first tag", "second tag"]
        root["folder_1"]["content_1"] = Content()
        root["folder_1"]["content_1"].tags = ["third tag"]
        root["folder_1"]["content_2"] = Content()
        root["folder_1"]["content_2"].tags = ["first tag", "third tag"]
        first_tag = Tag.query.filter(Tag.title == "first tag").one()
        assert [rel.name for rel in first_tag.items] == ["folder_1", "content_2"]
        second_tag = Tag.query.filter(Tag.title == "second tag").one()
        assert [rel.name for rel in second_tag.items] == ["folder_1"]
        third_tag = Tag.query.filter(Tag.title == "third tag").one()
        assert [rel.name for rel in third_tag.items] == ["content_1", "content_2"]

    def test_get_content_items_for_tag_title(self, root):
        from kotti.resources import Tag, TagsToContents, Content
        from kotti.views.util import content_with_tags

        root["folder_1"] = Content()
        root["folder_1"].tags = ["first tag", "second tag"]
        root["folder_1"]["content_1"] = Content()
        root["folder_1"]["content_1"].tags = ["third tag"]
        root["folder_1"]["content_2"] = Content()
        root["folder_1"]["content_2"].tags = ["first tag", "third tag"]

        result = (
            Content.query.join(TagsToContents)
            .join(Tag)
            .filter(Tag.title == "first tag")
            .all()
        )
        assert [res.name for res in result] == ["folder_1", "content_2"]
        result = (
            Content.query.join(TagsToContents)
            .join(Tag)
            .filter(Tag.title == "second tag")
            .all()
        )
        assert [res.name for res in result] == ["folder_1"]
        result = (
            Content.query.join(TagsToContents)
            .join(Tag)
            .filter(Tag.title == "third tag")
            .all()
        )
        assert sorted(res.name for res in result) == sorted(["content_1", "content_2"])

        # The same tests again, using content_with_tags():
        #
        #     About expected sort order:
        #
        #         In the first set of tests below, where we search by single
        #         tags, the query in the content_with_tags() function returns
        #         results in hierarchical order, from root.
        #
        # content_with_tags() is written to take a list of tags, but in the
        # default Kotti, presently, after some consideration about specialized
        # add-ons for searching, we do not support multiple tags searching, in
        # part to avoid establishing a specification.
        #
        result = content_with_tags(["first tag"])
        assert sorted([res.name for res in result]) == sorted(["folder_1", "content_2"])
        result = content_with_tags(["second tag"])
        assert sorted([res.name for res in result]) == sorted(["folder_1"])
        result = content_with_tags(["third tag"])
        assert sorted([res.name for res in result]) == sorted(
            ["content_1", "content_2"]
        )


class TestCommaSeparatedListWidget:
    def make_one(self):
        from kotti.views.form import CommaSeparatedListWidget

        return CommaSeparatedListWidget(template="")

    def test_widget_serialize_none(self):
        field = Mock()
        widget = self.make_one()
        widget.serialize(field, None)
        field.renderer.assert_called_with(widget.template, field=field, cstruct=[])

    def test_widget_serialize_null(self):
        field = Mock()
        widget = self.make_one()
        widget.serialize(field, colander.null)
        field.renderer.assert_called_with(widget.template, field=field, cstruct=[])

    def test_widget_serialize(self):
        field = Mock()
        widget = self.make_one()
        widget.serialize(field, ["yes"])
        field.renderer.assert_called_with(widget.template, field=field, cstruct=["yes"])

    def test_widget_deserialize_null(self):
        assert self.make_one().deserialize(None, colander.null) == colander.null

    def test_widget_deserialize(self):
        assert self.make_one().deserialize(None, "foo,bar") == ["foo", "bar"]



"""
Inheritance Diagram
-------------------

.. inheritance-diagram:: kotti.testing
"""
import os
from os.path import dirname
from os.path import join
from unittest import TestCase
from warnings import catch_warnings

import transaction
from pyramid import testing
from pyramid.events import NewResponse
from pyramid.interfaces import ILocation
from pyramid.security import ALL_PERMISSIONS
from pytest import mark
from zope.interface import implementer

# re-enable deprecation warnings during test runs
# however, let the `ImportWarning` produced by Babel's
# `localedata.py` vs `localedata/` show up once...
with catch_warnings():
    pass
    # import compiler
    # localedata, compiler    # make pyflakes happy... :p


# py.test markers (see http://pytest.org/latest/example/markers.html)
user = mark.user


BASE_URL = "http://localhost:6543"


class Dummy(dict):
    # noinspection PyMissingConstructor
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


class DummyRequest(testing.DummyRequest):
    is_xhr = False
    POST = dict()
    user = None
    referrer = None

    @staticmethod
    def is_response(ob):
        return (
            hasattr(ob, "app_iter")
            and hasattr(ob, "headerlist")
            and hasattr(ob, "status")
        )

    # noinspection PyPep8Naming
    @classmethod
    def blank(cls, path, environ=None, base_url=None, headers=None, POST=None, **kw):
        """
        ``request.blank`` is used in Kotti only when assigning slots, where
        the POST parameters are faked as a querystring.
        """
        from urllib.parse import parse_qsl

        def _decode(body):
            if not body:
                return {}
            return {x: y for x, y in parse_qsl(body)}

        if POST and isinstance(POST, bytes):
            POST = POST.decode()
        if POST and isinstance(POST, str):
            POST = _decode(POST)
        req = testing.DummyRequest(
            path=path, environ=environ, headers=headers, cookies=None, post=POST, **kw
        )
        return req


def asset(name):
    import kotti

    return open(join(dirname(kotti.__file__), "tests", name), "rb")


def includeme_login(config):
    """ Pyramid includeme hook.

    :param config: app config
    :type config: :class:`pyramid.config.Configurator`
    """

    config.add_view(login_view, name="login", renderer="kotti:templates/login.pt")


def includeme_layout(config):
    """ Pyramid includeme hook.

    :param config: app config
    :type config: :class:`pyramid.config.Configurator`
    """

    # override edit master layout with view master layout

    config.override_asset(
        to_override="kotti:templates/edit/master.pt",
        override_with="kotti:templates/view/master.pt",
    )


def login_view(request):
    return {}


def dummy_search(search_term, request):
    return "Not found. Sorry!"


def testing_db_url():
    return os.environ.get("KOTTI_TEST_DB_STRING", "sqlite://")


def _init_testing_db():
    from sqlalchemy import create_engine
    from kotti import get_settings
    from kotti.resources import initialize_sql

    database_url = testing_db_url()
    get_settings()["sqlalchemy.url"] = database_url
    session = initialize_sql(create_engine(database_url), drop_all=True)
    return session


def _populator():
    from kotti import DBSession
    from kotti.resources import Document
    from kotti.populate import populate

    populate()
    for doc in DBSession.query(Document)[1:]:
        DBSession.delete(doc)
    transaction.commit()


def _turn_warnings_into_errors():  # pragma: no cover
    # turn all warnings into errors, but let the `ImportWarning`
    # produced by Babel's `localedata.py` vs `localedata/` show up once...
    from babel import localedata

    localedata  # make pyflakes happy... :p
    from warnings import filterwarnings

    filterwarnings("error")


# noinspection PyPep8Naming
def setUp(init_db=True, **kwargs):
    # _turn_warnings_into_errors()

    from kotti import _resolve_dotted
    from kotti import conf_defaults

    tearDown()
    settings = conf_defaults.copy()
    settings["kotti.secret"] = "secret"
    settings["kotti.secret2"] = "secret2"
    settings["kotti.populators"] = "kotti.testing._populator"
    settings["pyramid_deform.tempdir"] = "/tmp"
    settings.update(kwargs.get("settings", {}))
    settings = _resolve_dotted(settings)
    kwargs["settings"] = settings
    config = testing.setUp(**kwargs)
    config.add_default_renderers()

    if init_db:
        _init_testing_db()

    transaction.begin()
    return config


# noinspection PyPep8Naming
def tearDown():
    from depot.manager import DepotManager
    from kotti import events
    from kotti import security
    from kotti.message import _inject_mailer

    # These should arguable use the configurator, so they don't need
    # to be torn down separately:
    events.clear()
    security.reset()

    _inject_mailer[:] = []
    transaction.abort()
    DepotManager._clear()
    testing.tearDown()


class UnitTestBase(TestCase):
    def setUp(self, **kwargs):
        self.config = setUp(**kwargs)

    def tearDown(self):
        tearDown()


class EventTestBase(TestCase):
    def setUp(self, **kwargs):
        super().setUp(**kwargs)
        self.config.include("kotti.events")


# Functional ----


def _functional_includeme(config):
    from kotti import DBSession

    def expire(event):
        DBSession.flush()
        DBSession.expire_all()

    config.add_subscriber(expire, NewResponse)


def _zope_testbrowser_pyquery(self):
    from pyquery import PyQuery

    return PyQuery(self.contents.replace('xmlns="http://www.w3.org/1999/xhtml', ""))


# noinspection PyPep8Naming
def setUpFunctional(global_config=None, **settings):
    from kotti import main
    from zope.testbrowser.wsgi import Browser
    from webtest import TestApp

    tearDown()

    _settings = {
        "sqlalchemy.url": testing_db_url(),
        "kotti.secret": "secret",
        "kotti.site_title": "Website des Kottbusser Tors",  # for mailing
        "kotti.populators": "kotti.testing._populator",
        "mail.default_sender": "kotti@localhost",
        "pyramid.includes": "kotti.testing._functional_includeme",
    }
    _settings.update(settings)

    host, port = BASE_URL.split(":")[-2:]
    app = main({}, **_settings)
    Browser.pyquery = property(_zope_testbrowser_pyquery)

    return dict(
        Browser=lambda: Browser(
            "http://{}:{}/".format(host[2:], int(port)), wsgi_app=app
        ),
        browser=Browser("http://{}:{}/".format(host[2:], int(port)), wsgi_app=app),
        test_app=TestApp(app),
    )


class FunctionalTestBase(TestCase):
    BASE_URL = BASE_URL

    def setUp(self, **kwargs):
        self.__dict__.update(setUpFunctional(**kwargs))

    def tearDown(self):
        tearDown()

    def login(self, login="admin", password="secret"):
        return self.test_app.post(
            "/@@login",
            {"login": login, "password": password, "submit": "submit"},
            status=302,
        )


@implementer(ILocation)
class RootFactory(dict):
    __name__ = ""  # root is required to have an empty name!
    __parent__ = None
    __acl__ = [("Allow", "role:admin", ALL_PERMISSIONS)]

    def __init__(self, request):
        super().__init__()


def dummy_view(context, request):
    return {}


def include_testing_view(config):
    """ Pyramid includeme hook.

    :param config: app config
    :type config: :class:`pyramid.config.Configurator`
    """

    config.add_view(
        dummy_view, context=RootFactory, renderer="kotti:tests/testing_view.pt"
    )

    config.add_view(
        dummy_view,
        name="secured",
        permission="view",
        context=RootFactory,
        renderer="kotti:tests/testing_view.pt",
    )


# noinspection PyPep8Naming
def setUpFunctionalStrippedDownApp(global_config=None, **settings):
    # An app that doesn't use Nodes at all
    _settings = {
        "kotti.base_includes": (
            "kotti kotti.views kotti.views.login kotti.views.users " "kotti.views.view"
        ),
        "kotti.use_tables": "principals",
        "kotti.populators": "kotti.populate.populate_users",
        "pyramid.includes": "kotti.testing.include_testing_view",
        "kotti.root_factory": "kotti.testing.RootFactory",
        "kotti.site_title": "My Stripped Down Kotti",
    }
    _settings.update(settings)

    return setUpFunctional(global_config, **_settings)


# noinspection PyPep8Naming
def registerDummyMailer():
    from pyramid_mailer.mailer import DummyMailer
    from kotti.message import _inject_mailer

    mailer = DummyMailer()
    _inject_mailer.append(mailer)
    return mailer


# set up deprecation warnings
from zope.deprecation.deprecation import deprecated  # noqa

for item in UnitTestBase, EventTestBase, FunctionalTestBase, _init_testing_db:
    name = getattr(item, "__name__", item)
    deprecated(
        name,
        "Unittest-style tests are deprecated as of Kotti 0.7. "
        "Please use pytest function arguments instead.",
    )



from pytest import fixture
from pytest import skip


# ``py.test --runslow`` causes the entire testsuite to be run, including test
# that are decorated with ``@@slow`` (scaffolding tests).
# see http://pytest.org/latest/example/simple.html#control-skipping-of-tests-according-to-command-line-option  # noqa


def pytest_addoption(parser):
    parser.addoption("--runslow", action="store_true", help="run slow tests")


def pytest_runtest_setup(item):
    if "slow" in item.keywords and not item.config.getoption("--runslow"):
        skip("need --runslow option to run")


# non-public test fixtures


@fixture
def app(db_session, setup_app):
    from webtest import TestApp

    return TestApp(setup_app)


@fixture
def extra_principals(db_session):
    """ Setup additional users 'bob', 'frank' and groups 'bobsgroup' and
    'franksgroup'.  Return the result of ``get_pricipals()``. """

    from kotti.security import get_principals

    P = get_principals()
    P["bob"] = dict(name="bob", title="Bob")
    P["frank"] = dict(name="frank", title="Frank")
    P["group:bobsgroup"] = dict(name="group:bobsgroup", title="Bob's Group")
    P["group:franksgroup"] = dict(name="group:franksgroup", title="Frank's Group")
    return P
